from .base_tree import *
from .py_acv import *
from .utils_exp import *
from .utils_sdp import *
import numpy as np
import cyext_acv, cyext_acv_nopa, cyext_acv_cache
from acv_explainers.utils import extend_partition
from sklearn.utils.validation import check_array, column_or_1d, check_consistent_length

class ACVTree(BaseTree):

    def shap_values(self, X, C=[[]], num_threads=10):
        """
        Estimate the Shapley Values of a set of samples using the Leaf estimator

        Args:
            X (numpy.ndarray): A matrix of samples (# samples X # features) on which to explain the model's output

            C (list[list[int]]): A list that contains a list of columns indices for each grouped variables

            num_threads (int): not used, deprecated

        Returns:
            shapley_values (numpy.ndarray): The Shapley Values of each sample (# samples X # features X # model's output)
        """
        X = check_array(X, dtype=[np.double])
        if not self.cache:
            return cyext_acv.shap_values_leaves_pa(X, self.data, self.values,
                                                   self.partition_leaves_trees,
                                                   self.leaf_idx_trees, self.leaves_nb, self.max_var,
                                                   self.node_idx_trees, C, num_threads)
        return self.shap_values_cache(X, C)

    def shap_values_acv_adap(self, X, S_star, N_star, size, C=[[]], num_threads=10):
        """
        Estimate the **Active** Shapley Values for a set of samples given **the corresponding** S_star and N_star of
        **each sample** using the Leaf estimator.

        Args:
            X (numpy.ndarray): A matrix of samples (# samples X # features) on which to explain the model's output

            S_star (numpy.ndarray): A matrix (# samples X # features) that contains the indices of the sufficient variables
                                  of each sample. For each row, the first elements should corresponds to the indices of
                                  the Sufficient variables.

            N_star (numpy.ndarray): A matrix (# samples X # features) that contains the indices of the null variables
                                  of each sample. For each row, the first elements should corresponds to the indices of
                                  the null variables.

            size (numpy.ndarray): A 1-D matrix (# samples) that contains the size of the corresponding sufficient variables
                                for each samples

            C (list[list[int]]): A list that contains a list of columns indices for each grouped variables

            num_threads (int): not used, deprecated

        Returns:
            active_shapley_values (numpy.ndarray): The **Active** Shapley Values of each sample (# samples X # features X # model's output)
        """
        X = check_array(X, dtype=[np.double])
        S_star = check_array(S_star, dtype=[np.long])
        N_star = check_array(N_star, dtype=[np.long])
        check_consistent_length(X, S_star)
        size = column_or_1d(size)
        return cyext_acv.shap_values_acv_leaves_adap(X, self.data, self.values,
                                                     self.partition_leaves_trees,
                                                     self.leaf_idx_trees, self.leaves_nb, self.max_var,
                                                     self.node_idx_trees, S_star, N_star, size, C, num_threads)

    def shap_values_acv(self, X, S_star, N_star, C=[[]], num_threads=10):
        """
        Estimate the **Active** Shapley Values for a set of samples given a **single** S_star and N_star **for all**
        the observations using the Leaf estimator.

        Args:
            X (numpy.ndarray): A matrix of samples (# samples X # features) on which to explain the model's output

            S_star (list): A list that contains the columns indices of the active variables

            N_star (list): A list that contains the columns indices of the null variables

            C (list[list[int]]): A list that contains a list of columns indices for each grouped variables

            num_threads (int): not used, deprecated

        Returns:
            numpy.ndarray: The **Active** Shapley Values of each sample (# samples X # features X # model's output)
        """
        X = check_array(X, dtype=[np.double])
        return cyext_acv.shap_values_acv_leaves(X, self.data, self.values,
                                                self.partition_leaves_trees,
                                                self.leaf_idx_trees, self.leaves_nb, self.max_var,
                                                self.node_idx_trees, S_star, N_star, C, num_threads)

    def importance_sdp_clf(self, X, data, C=[[]], pi_level=0.9, minimal=1, stop=True):
        """
        Estimate the Minimal-Sufficient Explanations of a set of samples for tree-based classifier models. It searches
        the Sufficient Explanations in the subspace of the 10-variables frequently selected in the tree-based model to
        reduce the complexity from 2**(features) to 2**(10).

        Args:
            X (numpy.ndarray): A matrix of samples (# samples X # features) on which to compute the Sufficient Explanations

            data (numpy.ndarray): The background dataset to use for the estimation of the explanations. It should be the
                                training samples.

            C (list[list[int]]): A list that contains a list of columns indices for each grouped variables

            pi_level (float): The minimal value of the Same Decision Probability (SDP) of the Sufficient Explanations

            minimal (int): It will search the Sufficient Explanations from subsets of size "minimal" instead of 1 by default

            stop (bool): If stop=True, it will stop searching for the Sufficient Explanations, if it does not find
                         any Sufficient Explanations smaller than (# features / 2), otherwise it will continues until
                         end.

        Returns:
            global_sdp_importance (numpy.ndarray): A 1-D matrix (# features) that is the global explanatory importance
                                                 based on samples X. For a given i, sdp_importance[i] corresponds to the
                                                 frequency of apparition of feature i in the Minimal Sufficient Explanations
                                                 of the set of samples X

            sdp_index (numpy.ndarray): A matrix (# samples X # features) that contains the column indices of the variables
                                       in the the Minimal Sufficient Explanations for each sample. For a given i, the positive
                                       value of sdp_index[i] corresponds to the Minimal Sufficient Explanations of observation i.

            size (numpy.ndarray): A 1-D matrix (# samples), size[i] corresponds to the size of the Minimal Sufficient Explanation
                                  of sample i.

            sdp (numpy.ndarray): A 1-D matrix (# samples) that contains the Same Decision Probability (SDP)
                                 of the Sufficient Explanation for each sample.


        """
        X = check_array(X, dtype=[np.double])
        data = check_array(data, dtype=[np.double])
        if X.shape[1] > 10:
            flat_list = [item for t in self.node_idx_trees for sublist in t for item in sublist]
            node_idx = pd.Series(flat_list)
            order_va = []
            for v in (node_idx.value_counts().keys()):
                order_va += [v]
            return self.importance_sdp_clf_search(X, data, C, pi_level, minimal, list(order_va[:10]), stop)
        else:
            return self.importance_sdp_clf_greedy(X, data, C, pi_level, minimal, stop=stop)

    def importance_sdp_clf_greedy(self, X, data, C=[[]], pi_level=0.9, minimal=1, stop=True):
        """
        Estimate the Minimal-Sufficient Explanations of a set of samples for tree-based classifier models. It searches
        the Sufficient Explanations in the space of all the variables thus it has a complexity of 2**(# features).

        Args:
            X (numpy.ndarray): A matrix of samples (# samples X # features) on which to compute the Sufficient Explanations

            data (numpy.ndarray): The background dataset to use for the estimation of the explanations. It should be the
                                training samples.

            C (list[list[int]]): A list that contains a list of columns indices for each grouped variables

            pi_level (float): The minimal value of the Same Decision Probability (SDP) of the Sufficient Explanations

            minimal (int): It will search the Sufficient Explanations from subsets of size "minimal" instead of 1 by default

            stop (bool): If stop=True, it will stop searching for the Sufficient Explanations, if it does not find
                         any Sufficient Explanations smaller than (# features / 2), otherwise it will continues until
                         end.

        Returns:
            global_sdp_importance (numpy.ndarray): A 1-D matrix (# features) that is the global explanatory importance
                                                 based on samples X. For a given i, sdp_importance[i] corresponds to the
                                                 frequency of apparition of feature i in the Minimal Sufficient Explanations
                                                 of the set of samples X

            sdp_index (numpy.ndarray): A matrix (# samples X # features) that contains the indices of the variables in the
                                     the Minimal Sufficient Explanations for each sample. For a given i, the positive
                                     value of sdp_index[i] corresponds to the Minimal Sufficient Explanations of
                                     observation i.

            size (numpy.ndarray): A 1-D matrix (# samples), size[i] corresponds to the size of the Minimal Sufficient Explanation
                                  of observation i

            sdp (numpy.ndarray): A 1-D matrix (# samples), sdp[i] that contains the Same Decision Probability (SDP)
                                 of the Sufficient Explanation of observation i
        """
        X = check_array(X, dtype=[np.double])
        data = check_array(data, dtype=[np.double])
        fX = np.argmax(self.model.predict_proba(X), axis=1).astype(np.long)
        y_pred = np.argmax(self.model.predict_proba(data), axis=1).astype(np.long)
        if safe_isinstance(self.model, ["xgboost.sklearn.XGBClassifier", "catboost.core.CatBoostClassifier",
                                        "lightgbm.sklearn.LGBMClassifier",
                                        "sklearn.ensemble.GradientBoostingClassifier",
                                        "sklearn.ensemble._gb.GradientBoostingClassifier",
                                        "sklearn.ensemble.gradient_boosting.GradientBoostingClassifier"
                                        ]) and \
                self.num_outputs == 1:
            return cyext_acv.global_sdp_clf(X, fX, y_pred, data, self.values_binary,
                                            self.partition_leaves_trees, self.leaf_idx_trees, self.leaves_nb,
                                            self.scalings, C, pi_level, minimal, stop)

        return cyext_acv.global_sdp_clf(X, fX, y_pred, data, self.values,
                                        self.partition_leaves_trees, self.leaf_idx_trees, self.leaves_nb,
                                        self.scalings, C, pi_level, minimal, stop)

    def importance_sdp_clf_search(self, X, data, C=[[]], pi_level=0.9, minimal=1, search_space=[], stop=True):
        """
        Estimate the Minimal-Sufficient Explanations of a set of samples for tree-based classifier models. It searches
        the Sufficient Explanations in the given subspace (search_space). Thus, it has a complexity of 2**(# search_space).

        Args:
            X (numpy.ndarray): A matrix of samples (# samples X # features) on which to compute the Sufficient Explanations

            data (numpy.ndarray): The background dataset to use for the estimation of the explanations. It should be the
                                training samples.

            C (list[list[int]]): A list that contains a list of columns indices for each grouped variables

            pi_level (float): The minimal value of the Same Decision Probability (SDP) of the Sufficient Explanations

            minimal (int): It will search the Sufficient Explanations from subsets of size "minimal" instead of 1 by default

            search_space (list): A list containing the variables (columns indices) on which to search the Sufficient Explanations.

            stop (bool): If stop=True, it will stop searching for the Sufficient Explanations, if it does not find
                         any Sufficient Explanations smaller than (# features / 2), otherwise it will continues until
                         end.

        Returns:
            global_sdp_importance (numpy.ndarray): A 1-D matrix (# features) that is the global explanatory importance
                                                 based on samples X. For a given i, sdp_importance[i] corresponds to the
                                                 frequency of apparition of feature i in the Minimal Sufficient Explanations
                                                 of the set of samples X

            sdp_index (numpy.ndarray): A matrix (# samples X # features) that contains the indices of the variables in the
                                     the Minimal Sufficient Explanations for each sample. For a given i, the positive
                                     value of sdp_index[i] corresponds to the Minimal Sufficient Explanations of
                                     observation i.

            size (numpy.ndarray): A 1-D matrix (# samples), size[i] corresponds to the size of the Minimal Sufficient Explanation
                                  of observation i

            sdp (numpy.ndarray): A 1-D matrix (# samples), sdp[i] that contains the Same Decision Probability (SDP)
                                 of the Sufficient Explanation of observation i
        """
        X = check_array(X, dtype=[np.double])
        data = check_array(data, dtype=[np.double])
        fX = np.argmax(self.model.predict_proba(X), axis=1).astype(np.long)
        y_pred = np.argmax(self.model.predict_proba(data), axis=1).astype(np.long)
        if safe_isinstance(self.model, ["xgboost.sklearn.XGBClassifier", "catboost.core.CatBoostClassifier",
                                        "lightgbm.sklearn.LGBMClassifier",
                                        "sklearn.ensemble.GradientBoostingClassifier",
                                        "sklearn.ensemble._gb.GradientBoostingClassifier",
                                        "sklearn.ensemble.gradient_boosting.GradientBoostingClassifier"]) and \
                self.num_outputs == 1:
            return cyext_acv.global_sdp_clf_approx(np.array(X, dtype=np.float), fX, y_pred, data, self.values_binary,
                                                   self.partition_leaves_trees, self.leaf_idx_trees, self.leaves_nb,
                                                   self.scalings, C, pi_level, minimal, search_space, stop)

        return cyext_acv.global_sdp_clf_approx(np.array(X, dtype=np.float), fX, y_pred, data, self.values,
                                               self.partition_leaves_trees, self.leaf_idx_trees, self.leaves_nb,
                                               self.scalings, C, pi_level, minimal, search_space, stop)

    def compute_sdp_clf(self, X, S, data, num_threads=10):
        """
        Estimate the Same Decision Probability (SDP) of a set of samples X given subset S using the Leaf estimator.

        Args:
            X (numpy.ndarray): A matrix of samples (# samples X # features) on which to compute the SDP

            S (numpy.ndarray): A 1-D that contains the indices of the variable on which we condition to compute the SDP

            data (numpy.ndarray): The background dataset to use for the estimation of the SDP. It should be the
                                  training samples.

            num_threads (int): not used, deprecated

        Returns:
            sdp (numpy.ndarray)  A 1-D matrix (# samples), sdp[i] that contains the Same Decision Probability (SDP)
                                 of the Sufficient Explanation of observation i
        """
        X = check_array(X, dtype=[np.double])
        data = check_array(data, dtype=[np.double])
        S = column_or_1d(S)
        fX = np.argmax(self.model.predict_proba(X), axis=1).astype(np.long)
        y_pred = np.argmax(self.model.predict_proba(data), axis=1).astype(np.long)
        if safe_isinstance(self.model, ["xgboost.sklearn.XGBClassifier", "catboost.core.CatBoostClassifier",
                                        "lightgbm.sklearn.LGBMClassifier",
                                        "sklearn.ensemble.GradientBoostingClassifier",
                                        "sklearn.ensemble._gb.GradientBoostingClassifier",
                                        "sklearn.ensemble.gradient_boosting.GradientBoostingClassifier"]) and \
                self.num_outputs == 1:
            return cyext_acv.compute_sdp_clf(np.array(X, dtype=np.float), fX, y_pred, S, data, self.values_binary,
                                             self.partition_leaves_trees,
                                             self.leaf_idx_trees, self.leaves_nb, self.scalings, num_threads)

        return cyext_acv.compute_sdp_clf(np.array(X, dtype=np.float), fX, y_pred, S, data, self.values,
                                         self.partition_leaves_trees,
                                         self.leaf_idx_trees, self.leaves_nb, self.scalings, num_threads)

    def compute_sdp_clf_cat(self, X, S, data, num_threads=10):
        """
        Estimate the Same Decision Probability (SDP) of a set of samples X given subset S using the Discrete estimator.

        Args:
            X (numpy.ndarray): A matrix of samples (# samples X # features) on which to compute the SDP

            S (numpy.ndarray): A 1-D that contains the indices of the variable on which we condition to compute the SDP

            data (numpy.ndarray): The background dataset to use for the estimation of the SDP. It should be the
                                training samples.

            num_threads (int): not used, deprecated

        Returns:
            sdp (numpy.ndarray)  A 1-D matrix (# samples), sdp[i] that contains the Same Decision Probability (SDP)
                                 of the Sufficient Explanation of observation i
        """
        X = check_array(X, dtype=[np.double])
        data = check_array(data, dtype=[np.double])
        S = column_or_1d(S)
        fX = np.argmax(self.model.predict_proba(X), axis=1).astype(np.long)
        y_pred = np.argmax(self.model.predict_proba(data), axis=1).astype(np.long)
        if safe_isinstance(self.model, ["xgboost.sklearn.XGBClassifier", "catboost.core.CatBoostClassifier",
                                        "lightgbm.sklearn.LGBMClassifier",
                                        "sklearn.ensemble.GradientBoostingClassifier",
                                        "sklearn.ensemble._gb.GradientBoostingClassifier",
                                        "sklearn.ensemble.gradient_boosting.GradientBoostingClassifier"]) and \
                self.num_outputs == 1:
            return cyext_acv.compute_sdp_clf_cat(np.array(X, dtype=np.float), fX, y_pred, S, data, self.values_binary,
                                                 self.partition_leaves_trees,
                                                 self.leaf_idx_trees, self.leaves_nb, self.scalings, num_threads)

        return cyext_acv.compute_sdp_clf_cat(np.array(X, dtype=np.float), fX, y_pred, S, data, self.values,
                                             self.partition_leaves_trees,
                                             self.leaf_idx_trees, self.leaves_nb, self.scalings, num_threads)

    # ----------------------------------------------------RELATED FUNCTION---------------------------------------------#

    def shap_values_nopa(self, X, C=[[]], num_threads=10):
        """
                Same as **shap_values** function, but does not use parallelism
        """
        X = check_array(X, dtype=[np.double])
        return cyext_acv_nopa.shap_values_leaves_nopa(np.array(X, dtype=np.float), self.data, self.values,
                                                      self.partition_leaves_trees,
                                                      self.leaf_idx_trees, self.leaves_nb, self.max_var,
                                                      self.node_idx_trees, C, num_threads)

    def shap_values_acv_nopa(self, X, S_star, N_star, C=[[]], num_threads=10):
        """
                Same as **shap_values_acv** function, but does not use parallelism
        """
        X = check_array(X, dtype=[np.double])
        return cyext_acv_nopa.shap_values_acv_leaves_nopa(np.array(X, dtype=np.float), self.data, self.values,
                                                          self.partition_leaves_trees,
                                                          self.leaf_idx_trees, self.leaves_nb, self.max_var,
                                                          self.node_idx_trees, S_star, N_star, C, num_threads)

    def shap_values_acv_adap_nopa(self, X, S_star, N_star, size, C=[[]], num_threads=10):
        """
                Same as **shap_values_acv_adap** function, but does not use parallelism
        """
        X = check_array(X, dtype=[np.double])
        S_star = check_array(S_star, dtype=[np.long])
        N_star = check_array(N_star, dtype=[np.long])
        check_consistent_length(X, S_star)
        size = column_or_1d(size)
        return cyext_acv_nopa.shap_values_acv_leaves_adap_nopa(np.array(X, dtype=np.float), self.data, self.values,
                                                               self.partition_leaves_trees,
                                                               self.leaf_idx_trees, self.leaves_nb, self.max_var,
                                                               self.node_idx_trees, S_star, N_star, size, C,
                                                               num_threads)

    def shap_values_cache(self, X, C=[[]], num_threads=10):
        """
                Same as **shap_values**, but use cached values to speed up computation

        """
        X = check_array(X, dtype=[np.double])
        return cyext_acv_cache.shap_values_leaves_cache(np.array(X, dtype=np.float), self.data, self.values,
                                                        self.partition_leaves_trees,
                                                        self.leaf_idx_trees, self.leaves_nb, self.lm, self.lm_s,
                                                        self.lm_si,
                                                        self.max_var,
                                                        self.node_idx_trees, C, num_threads)

    def shap_values_cache_nopa(self, X, C=[[]], num_threads=10):
        """
                Same as **shap_values** but use cached values and no parallelism
        """
        X = check_array(X, dtype=[np.double])
        return cyext_acv_nopa.shap_values_leaves_cache_nopa(np.array(X, dtype=np.float), self.data, self.values,
                                                            self.partition_leaves_trees,
                                                            self.leaf_idx_trees, self.leaves_nb, self.lm, self.lm_s,
                                                            self.lm_si,
                                                            self.max_var,
                                                            self.node_idx_trees, C, num_threads)

    def importance_sdp_clf_nopa(self, X, data, C=[[]], pi_level=0.9, minimal=0):
        """
                Same as importance_sdp_clf, but does not use parallelism
        """
        X = check_array(X, dtype=[np.double])
        data = check_array(data, dtype=[np.double])
        fX = np.argmax(self.model.predict_proba(X), axis=1).astype(np.long)
        y_pred = np.argmax(self.model.predict_proba(data), axis=1).astype(np.long)
        return cyext_acv_nopa.global_sdp_clf_nopa(X, fX, y_pred, data, self.values,
                                                  self.partition_leaves_trees, self.leaf_idx_trees, self.leaves_nb,
                                                  self.scalings, C, pi_level, minimal)

    def compute_sdp_clf_nopa(self, X, S, data, num_threads=10):
        """
                Same as compute_sdp_clf, but does not use parallelism
        """
        X = check_array(X, dtype=[np.double])
        data = check_array(data, dtype=[np.double])
        S = column_or_1d(S)
        fX = np.argmax(self.model.predict_proba(X), axis=1).astype(np.long)
        y_pred = np.argmax(self.model.predict_proba(data), axis=1).astype(np.long)
        return cyext_acv_nopa.compute_sdp_clf_nopa(np.array(X, dtype=np.float), fX, y_pred, S, data, self.values,
                                                   self.partition_leaves_trees,
                                                   self.leaf_idx_trees, self.leaves_nb, self.scalings, num_threads)

    def compute_sdp_clf_cat_nopa(self, X, S, data, num_threads=10):
        """
                Same as compute_sdp_clf_cat, but does not use parallelism
        """
        X = check_array(X, dtype=[np.double])
        S = column_or_1d(S)
        fX = np.argmax(self.model.predict_proba(X), axis=1).astype(np.long)
        y_pred = np.argmax(self.model.predict_proba(data), axis=1).astype(np.long)
        return cyext_acv_nopa.compute_sdp_clf_cat_nopa(np.array(X, dtype=np.float), fX, y_pred, S, data, self.values,
                                                       self.partition_leaves_trees,
                                                       self.leaf_idx_trees, self.leaves_nb, self.scalings, num_threads)

    # ------------------------------------------------PYTHON VERSION---------------------------------------------------#

    def py_shap_values(self, x, C=[[]]):
        out = np.zeros((x.shape[0], x.shape[1], self.num_outputs))
        for i in range(len(self.trees)):
            out += shap_values_leaves(x, self.partition_leaves_trees[i], self.data,
                                      self.node_idx_trees[i],
                                      self.leaf_idx_trees[i], self.leaves_nb[i], self.node_sample_weight[i],
                                      self.values[i], C,
                                      self.num_outputs)
        return out

    def py_shap_valuesv2(self, x, C=[[]]):
        out = np.zeros((x.shape[0], x.shape[1], self.num_outputs))
        for i in range(len(self.trees)):
            out += shap_values_leaves_v2(x, self.partition_leaves_trees[i], self.data,
                                         self.node_idx_trees[i],
                                         self.leaf_idx_trees[i], self.leaves_nb[i], self.node_sample_weight[i],
                                         self.values[i], C,
                                         self.num_outputs)
        return out

    def py_shap_values_acv(self, x, S_star, N_star, C=[[]]):
        out = np.zeros((x.shape[0], x.shape[1], self.num_outputs))
        for i in range(len(self.trees)):
            out += shap_values_acv_leaves(x, self.partition_leaves_trees[i], self.data,
                                          self.node_idx_trees[i],
                                          self.leaf_idx_trees[i], self.leaves_nb[i], self.node_sample_weight[i],
                                          self.values[i], C, S_star,
                                          N_star,
                                          self.num_outputs)
        return out

    def py_compute_sdp_reg(self, X, tX, S, data):
        return compute_sdp_reg(X, tX, self, S, data=data)

    def py_compute_sdp_clf(self, X, S, data, tX=0):
        return compute_sdp_clf(X, tX, self, S, data=data)

    def py_compute_sdp_reg_cat(self, X, tX, S, data):
        return compute_sdp_reg_cat(X, tX, model=self, S=S, data=data)

    def py_compute_sdp_clf_cat(self, X, S, data, tX=0):
        return compute_sdp_clf_cat(X, tX, model=self, S=S, data=data)

    def py_compute_exp(self, X, S, data):
        return compute_exp(X=X, model=self, S=S, data=data)

    def py_compute_exp_cat(self, X, S, data):
        return compute_exp_cat(X=X, model=self, S=S, data=data)

    def py_compute_local_sdp_clf(self, X, threshold, proba, index, data, final_coal, decay, C, verbose):
        return local_sdp(X, threshold, proba, index, data, final_coal, decay, C, verbose,
                         self.compute_sdp_clf)

    def py_compute_local_sdp_reg(self, X, threshold, proba, index, data, final_coal, decay, C, verbose):
        return local_sdp(X, threshold, proba, index, data, final_coal, decay, C, verbose, self.compute_sdp_reg)

    def py_swing_values_clf(self, X, S, data, threshold, tX=0):
        return np.array(self.compute_sdp_clf(X=X, tX=tX, S=S, data=data) >= threshold, dtype=float)

    def py_swing_values_reg(self, X, tX, S, data, threshold):
        return np.array(self.compute_sdp_reg(X=X, tX=tX, S=S, data=data) >= threshold, dtype=float)

    def py_swing_sv_clf(self, X, data, threshold, C, tX=0):
        return swing_tree_shap(X, tX, threshold, data, C, self.swing_values_clf)

    def py_swing_sv_reg(self, X, data, threshold, C, tX=0):
        return swing_tree_shap(X, tX, threshold, data, C, self.swing_values_reg)

    def py_global_sdp_importance_clf(self, data, data_bground, columns_names, pi_level, decay, threshold,
                                     proba, C, verbose):

        return global_sdp_importance(data, data_bground, columns_names, pi_level, decay, threshold,
                                     proba, C, verbose, self.compute_sdp_clf)

    def py_global_sdp_importance_reg(self, data, data_bground, columns_names, pi_level, decay, threshold,
                                     proba, C, verbose):
        return global_sdp_importance(data, data_bground, columns_names, pi_level, decay, threshold,
                                     proba, C, verbose, self.compute_sdp_reg)

    def py_shap_values_notoptimized(self, X, data, C=[[]]):
        return shap_values_leaves_notoptimized(X, data, C, self)

    def py_shap_values_discrete_notoptimized(self, X, data, C=[[]]):
        return shap_values_discrete_notoptimized(X, data, C, self)

    # ------------------------------------------------FOR DEVELOPMENT--------------------------------------------------#

    def importance_sdp_clf_ptrees(self, X, data, C=[[]], pi_level=0.9, minimal=0, stop=True):
        fX = np.argmax(self.model.predict_proba(X), axis=1).astype(np.long)
        y_pred = np.argmax(self.model.predict_proba(data), axis=1).astype(np.long)
        if safe_isinstance(self.model, ["xgboost.sklearn.XGBClassifier", "catboost.core.CatBoostClassifier",
                                        "lightgbm.sklearn.LGBMClassifier",
                                        "sklearn.ensemble.GradientBoostingClassifier",
                                        "sklearn.ensemble._gb.GradientBoostingClassifier",
                                        "sklearn.ensemble.gradient_boosting.GradientBoostingClassifier"]) and \
                self.num_outputs == 1:
            return cyext_acv.global_sdp_clf_ptrees(np.array(X, dtype=np.float), fX, y_pred, data, self.values_binary,
                                                   self.partition_leaves_trees, self.leaf_idx_trees, self.leaves_nb,
                                                   self.scalings, C, pi_level, minimal, stop)

        return cyext_acv.global_sdp_clf_ptrees(np.array(X, dtype=np.float), fX, y_pred, data, self.values,
                                               self.partition_leaves_trees, self.leaf_idx_trees, self.leaves_nb,
                                               self.scalings, C, pi_level, minimal, stop)

    def compute_sdp_reg(self, X, tX, S, data, num_threads=10):
        # if self.partition_leaves_trees.shape[0] > 1:
        #     raise NotImplementedError('Continuous SDP is currently available only for trees with n_trees=1')
        fX = self.predict(X)
        y_pred = self.predict(data)
        return cyext_acv.compute_sdp_reg(np.array(X, dtype=np.float), fX, tX, y_pred, S, data, self.values,
                                         self.partition_leaves_trees,
                                         self.leaf_idx_trees, self.leaves_nb, self.scalings, num_threads)

    def compute_sdp_reg_cat(self, X, tX, S, data, num_threads=10):
        # raise Warning('The current implementation may take a long time if n_trees and depth are large. The number of '
        #               'operation is 2**(depth*n_trees)')
        fX = self.predict(X)
        y_pred = self.predict(data)
        return cyext_acv.compute_sdp_reg_cat(np.array(X, dtype=np.float), fX, tX, y_pred, S, data, self.values,
                                             self.partition_leaves_trees,
                                             self.leaf_idx_trees, self.leaves_nb, self.scalings, num_threads)

    def swing_sv_clf(self, X, data, C=[[]], thresholds=0.9, num_threads=10):
        fX = np.argmax(self.model.predict_proba(X), axis=1).astype(np.long)
        y_pred = np.argmax(self.model.predict_proba(data), axis=1).astype(np.long)
        return cyext_acv.swing_sv_clf_direct(np.array(X, dtype=np.float), fX, y_pred, data, self.values,
                                             self.partition_leaves_trees,
                                             self.leaf_idx_trees, self.leaves_nb, self.scalings, C, thresholds,
                                             num_threads)

    def swing_sv_clf_nopa(self, X, data, C=[[]], thresholds=0.9, num_threads=10):
        fX = np.argmax(self.model.predict_proba(X), axis=1).astype(np.long)
        y_pred = np.argmax(self.model.predict_proba(data), axis=1).astype(np.long)
        return cyext_acv_nopa.swing_sv_clf_direct_nopa(np.array(X, dtype=np.float), fX, y_pred, data, self.values,
                                                       self.partition_leaves_trees,
                                                       self.leaf_idx_trees, self.leaves_nb, self.scalings, C,
                                                       thresholds, num_threads)

    def demo_swing_sv_clf(self, X, data, C):
        return swing_tree_shap_clf(X, data, C, self.compute_sdp_clf)

    def demo_swing_sv_clf_cat(self, X, data, C):
        return swing_tree_shap_clf(X, data, C, self.compute_sdp_clf_cat)

    def swing_sv_clf_slow(self, X, data, C=[[]], thresholds=0.9, num_threads=5):
        fX = np.argmax(self.model.predict_proba(X), axis=1).astype(np.long)
        y_pred = np.argmax(self.model.predict_proba(data), axis=1).astype(np.long)
        return cyext_acv.swing_sv_clf(np.array(X, dtype=np.float), fX, y_pred, data, self.values,
                                      self.partition_leaves_trees, self.leaf_idx_trees, self.leaves_nb, self.scalings,
                                      C, thresholds, num_threads)

    def importance_sdp_reg_cat(self, X, tX, data, C=[[]], pi_level=0.9, minimal=0, stop=True):
        fX = self.predict(X)
        y_pred = self.predict(data)
        return cyext_acv.global_sdp_reg_cat(np.array(X, dtype=np.float), fX, tX, y_pred, data, self.values,
                                            self.partition_leaves_trees,
                                            self.leaf_idx_trees, self.leaves_nb, self.scalings, C, pi_level,
                                            minimal, stop)

    def importance_sdp_reg(self, X, tX, data, C=[[]], pi_level=0.9, minimal=0, stop=True):
        # if self.partition_leaves_trees.shape[0] > 1:
        #     raise NotImplementedError('Continuous SDP is currently available only for trees with n_trees=1')
        fX = self.predict(X)
        y_pred = self.predict(data)
        return cyext_acv.global_sdp_reg(np.array(X, dtype=np.float), fX, tX, y_pred, data, self.values,
                                        self.partition_leaves_trees, self.leaf_idx_trees, self.leaves_nb,
                                        self.scalings, C, pi_level, minimal, stop)

    def importance_sdp_reg_cat_nopa(self, X, tX, data, C=[[]], pi_level=0.9, minimal=0):
        fX = self.predict(X)
        y_pred = self.predict(data)
        return cyext_acv_nopa.global_sdp_reg_cat_nopa(np.array(X, dtype=np.float), fX, tX, y_pred, data, self.values,
                                                      self.partition_leaves_trees,
                                                      self.leaf_idx_trees, self.leaves_nb, self.scalings, C,
                                                      pi_level, minimal)

    def importance_sdp_reg_nopa(self, X, tX, data, C=[[]], pi_level=0.9, minimal=0):
        # if self.partition_leaves_trees.shape[0] > 1:
        #     raise NotImplementedError('Continuous SDP is currently available only for trees with n_trees=1')
        fX = self.predict(X)
        y_pred = self.predict(data)
        return cyext_acv_nopa.global_sdp_reg_nopa(np.array(X, dtype=np.float), fX, tX, y_pred, data, self.values,
                                                  self.partition_leaves_trees, self.leaf_idx_trees, self.leaves_nb,
                                                  self.scalings, C, pi_level, minimal)

    def compute_exp(self, X, S, data, num_threads=10):
        return cyext_acv.compute_exp(np.array(X, dtype=np.float), S, data, self.values, self.partition_leaves_trees,
                                     self.leaf_idx_trees, self.leaves_nb, self.scalings, num_threads)

    def compute_exp_cat(self, X, S, data, num_threads=10):
        return cyext_acv.compute_exp_cat(np.array(X, dtype=np.float), S, data, self.values, self.partition_leaves_trees,
                                         self.leaf_idx_trees, self.leaves_nb, self.scalings,
                                         num_threads)

    def compute_exp_normalized(self, X, S, data, num_threads=10):
        return cyext_acv.compute_exp_normalized(np.array(X, dtype=np.float), S, data, self.values,
                                                self.partition_leaves_trees,
                                                self.leaf_idx_trees, self.leaves_nb, self.scalings,
                                                num_threads)

    def compute_exp_normalized_nopa(self, X, S, data, num_threads=10):
        return cyext_acv.compute_exp_normalized_nopa(np.array(X, dtype=np.float), S, data, self.values,
                                                     self.partition_leaves_trees,
                                                     self.leaf_idx_trees, self.leaves_nb, self.scalings,
                                                     num_threads)

    def shap_values_normalized_cache(self, X, C=[[]], num_threads=10):
        return cyext_acv_cache.shap_values_leaves_normalized_cache(np.array(X, dtype=np.float), self.data, self.values,
                                                                   self.partition_leaves_trees,
                                                                   self.leaf_idx_trees, self.leaves_nb, self.lm_n,
                                                                   self.lm_s_n, self.lm_si_n,
                                                                   self.max_var,
                                                                   self.node_idx_trees, C, num_threads)

    def shap_values_normalized_cache_nopa(self, X, C=[[]], num_threads=10):
        return cyext_acv_nopa.shap_values_leaves_normalized_cache_nopa(np.array(X, dtype=np.float), self.data,
                                                                       self.values,
                                                                       self.partition_leaves_trees,
                                                                       self.leaf_idx_trees, self.leaves_nb, self.lm_n,
                                                                       self.lm_s_n, self.lm_si_n,
                                                                       self.max_var,
                                                                       self.node_idx_trees, C, num_threads)

    def shap_values_normalized(self, X, C=[[]], num_threads=10):
        if not self.cache_normalized:
            return cyext_acv.shap_values_leaves_normalized(np.array(X, dtype=np.float), self.data, self.values,
                                                           self.partition_leaves_trees,
                                                           self.leaf_idx_trees, self.leaves_nb, self.max_var,
                                                           self.node_idx_trees, C, num_threads)
        return self.shap_values_normalized_cache(X, C)

    def leaves_cache(self, C=[[]], num_threads=10):
        return cyext_acv_cache.leaves_cache(self.data, self.values, self.partition_leaves_trees,
                                            self.leaf_idx_trees, self.leaves_nb, self.max_var,
                                            self.node_idx_trees, C, num_threads)

    def leaves_cache_normalized(self, C=[[]], num_threads=10):
        return cyext_acv_cache.leaves_cache_normalized(self.data, self.values, self.partition_leaves_trees,
                                                       self.leaf_idx_trees, self.leaves_nb, self.max_var,
                                                       self.node_idx_trees, C, num_threads)

    def leaves_cache_nopa(self, C=[[]], num_threads=10):
        return cyext_acv_nopa.leaves_cache_nopa(self.data, self.values, self.partition_leaves_trees,
                                                self.leaf_idx_trees, self.leaves_nb, self.max_var,
                                                self.node_idx_trees, C, num_threads)

    def leaves_cache_normalized_nopa(self, C=[[]], num_threads=10):
        return cyext_acv_nopa.leaves_cache_normalized_nopa(self.data, self.values, self.partition_leaves_trees,
                                                           self.leaf_idx_trees, self.leaves_nb, self.max_var,
                                                           self.node_idx_trees, C, num_threads)

    def compute_exp_nopa(self, X, S, data, num_threads=10):
        return cyext_acv_nopa.compute_exp_nopa(np.array(X, dtype=np.float), S, data, self.values,
                                               self.partition_leaves_trees,
                                               self.leaf_idx_trees, self.leaves_nb, self.scalings, num_threads)

    def compute_exp_cat_nopa(self, X, S, data, num_threads=10):
        return cyext_acv_nopa.compute_exp_cat_nopa(np.array(X, dtype=np.float), S, data, self.values,
                                                   self.partition_leaves_trees,
                                                   self.leaf_idx_trees, self.leaves_nb, self.scalings,
                                                   num_threads)

    def compute_sdp_reg_nopa(self, X, tX, S, data, num_threads=10):
        # if self.partition_leaves_trees.shape[0] > 1:
        #     raise NotImplementedError('Continuous SDP is currently available only for trees with n_trees=1')
        fX = self.predict(X)
        y_pred = self.predict(data)
        return cyext_acv_nopa.compute_sdp_reg_nopa(np.array(X, dtype=np.float), fX, tX, y_pred, S, data, self.values,
                                                   self.partition_leaves_trees,
                                                   self.leaf_idx_trees, self.leaves_nb, self.scalings, num_threads)

    def compute_sdp_reg_cat_nopa(self, X, tX, S, data, num_threads=10):
        # raise Warning('The current implementation may take a long time if n_trees > 10 and depth > 6')
        fX = self.predict(X)
        y_pred = self.predict(data)
        return cyext_acv_nopa.compute_sdp_reg_cat_nopa(np.array(X, dtype=np.float), fX, tX, y_pred, S, data,
                                                       self.values, self.partition_leaves_trees,
                                                       self.leaf_idx_trees, self.leaves_nb, self.scalings, num_threads)

    def compute_msdp_clf(self, X, S, data, model=None, N=10000):
        """
        Compute marginal SDP of a set of sample X given S for classifier
        """
        # if data:
        #     return msdp_true(X, S, self.model, self.data)

        if model == None:
            return msdp(X, S, self.model, data)
        return msdp(X, S, model, data)

    def importance_msdp_clf_search(self, X, data, model=None, C=[[]], minimal=1, pi_level=0.9, r_search_space=None,
                                   stop=True):
        """
        Compute the marginal minimal sufficient explanations of any classifier
        """
        # if data:
        #     return importance_msdp_clf_true(X.values, self.model, self.data, C=C,
        #                                          minimal=minimal,
        #                                          pi_level=pi_level)
        if model == None:
            return importance_msdp_clf_search(X=X, rg_data=data, model=self.model, C=C, minimal=minimal,
                                              pi_level=pi_level, r_search_space=r_search_space, stop=stop)
        return importance_msdp_clf_search(X=X, rg_data=data, model=model, C=C, minimal=minimal,
                                          pi_level=pi_level, r_search_space=r_search_space, stop=stop)

    def importance_msdp_clf(self, X, data, model=None, C=[[]], pi_level=0.9, minimal=1, stop=True):
        if X.shape[1] > 15:
            flat_list = [item for t in self.node_idx_trees for sublist in t for item in sublist]
            node_idx = pd.Series(flat_list)
            order_va = []
            for v in (node_idx.value_counts().keys()):
                order_va += [v]
            return self.importance_msdp_clf_search(X=X, data=data, model=model, C=C, pi_level=pi_level,
                                                   minimal=minimal,
                                                   r_search_space=order_va[:15], stop=stop)
        else:
            return self.importance_msdp_clf_search(X=X, data=data, model=model, C=C, pi_level=pi_level,
                                                   minimal=minimal, stop=stop)

    def compute_msdp_reg(self, X, S, data, model=None, N=10000, threshold=0.2):
        """
         Compute marginal SDP of a set of sample X given S for regressor
        """
        # if data:
        #     return msdp_true(X, S, self.model, self.data)

        if model == None:
            return msdp_reg(X, S, self.model, data, threshold)
        return msdp_reg(X, S, model, data, threshold)

    def importance_msdp_reg_search(self, X, data, model=None, C=[[]], minimal=1, pi_level=0.9, threshold=0.2,
                                   r_search_space=None, stop=True):
        """
        Compute the marginal minimal sufficient explanations of any regressor model
        """
        # if data:
        #     return importance_msdp_clf_true(X.values, self.model, self.data, C=C,
        #                                          minimal=minimal,
        #                                          pi_level=pi_level)
        if model == None:
            return importance_msdp_reg_search(X, rg_data=data, model=self.model, C=C, minimal=minimal,
                                              pi_level=pi_level, threshold=threshold,
                                              r_search_space=r_search_space, stop=stop)
        return importance_msdp_reg_search(X, rg_data=data, model=model, C=C, minimal=minimal, pi_level=pi_level,
                                          threshold=threshold, r_search_space=r_search_space, stop=stop)

    def importance_msdp_reg(self, X, data, model=None, C=[[]], pi_level=0.9, minimal=1, threshold=0.2, stop=True):
        """
        Compute the marginal minimal sufficient explanations of any regressor model
        """
        if X.shape[1] > 15:
            flat_list = [item for t in self.node_idx_trees for sublist in t for item in sublist]
            node_idx = pd.Series(flat_list)
            order_va = []
            for v in (node_idx.value_counts().keys()):
                order_va += [v]
            return self.importance_msdp_reg_search(X=X, data=data, model=model, C=C, pi_level=pi_level,
                                                   minimal=minimal,
                                                   r_search_space=order_va[:15], threshold=threshold, stop=stop)
        else:
            return self.importance_msdp_reg_search(X=X, data=data, model=model, C=C, pi_level=pi_level,
                                                   minimal=minimal, threshold=threshold, stop=stop)

    def compute_exp_shaff(self, X, data, y_data, S, min_node_size=5):
        exp = np.zeros((X.shape[0], 1))
        for i in range(X.shape[0]):
            exp[i] = compute_shaff_exp(X=X[i], S=S, model=self, data=data, Y=y_data, min_node_size=min_node_size)
        return exp

    def compute_sdp_clf_shaff(self, X, y_X, data, y_data, S, min_node_size=5):
        sdp = np.zeros((X.shape[0], 1))
        for i in range(X.shape[0]):
            sdp[i] = compute_shaff_sdp_clf(X=X[i], y_X=y_X[i], S=S, model=self, data=data, Y=y_data,
                                           min_node_size=min_node_size)
        return sdp

    def compute_sdp_shaff(self, X, y_X, t, data, y_data, S, min_node_size=5):
        sdp = np.zeros((X.shape[0], 1))
        for i in range(X.shape[0]):
            sdp[i] = compute_shaff_sdp(X=X[i], t=t, y_X=y_X[i], S=S, model=self, data=data, Y=y_data,
                                       min_node_size=min_node_size)
        return sdp

    def compute_quantile_shaff(self, X, data, y_data, S, min_node_size=5, quantile=95):
        exp = np.zeros((X.shape[0], 1))
        for i in range(X.shape[0]):
            exp[i] = compute_shaff_quantile(X=X[i], S=S, model=self, data=data, Y=y_data, min_node_size=min_node_size,
                                            quantile=quantile)
        return exp

    def compute_sdp_rf(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20):
        sdp = cyext_acv.compute_sdp_rf(x, y, data, y_data, S, self.features, self.thresholds, self.children_left,
                                       self.children_right, self.max_depth, min_node_size, classifier, t)
        return sdp

    def compute_cdf_rf(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20):
        sdp = cyext_acv.compute_cdf_rf(x, y, data, y_data, S, self.features, self.thresholds, self.children_left,
                                       self.children_right, self.max_depth, min_node_size, classifier, t)
        return sdp

    def importance_sdp_rf(self, x, y, data, y_data, min_node_size=5, classifier=1, t=20,
                          C=[[]], pi_level=0.9, minimal=1, stop=True):

        if x.shape[1] > 10:

            flat_list = [item for t in self.node_idx_trees for sublist in t for item in sublist]
            node_idx = pd.Series(flat_list)
            search_space = []
            for v in (node_idx.value_counts().keys()):
                search_space += [v]
        else:
            search_space = [i for i in range(x.shape[1])]

        sdp = cyext_acv.global_sdp_rf(x, y, data, y_data, self.features, self.thresholds, self.children_left,
                                      self.children_right, self.max_depth, min_node_size, classifier, t, C,
                                      pi_level, minimal, stop, search_space[:10])
        return sdp

    def compute_exp_rf(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20):
        exp = cyext_acv.compute_exp_rf(x, y, data, y_data, S, self.features, self.thresholds, self.children_left,
                                       self.children_right, self.max_depth, min_node_size, classifier, t)
        return exp

    def compute_quantile_rf(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20, quantile=95):
        y_quantiles = cyext_acv.compute_quantile_rf(x, y, data, y_data, S, self.features, self.thresholds,
                                                    self.children_left,
                                                    self.children_right, self.max_depth, min_node_size, classifier, t,
                                                    quantile)
        return y_quantiles

    def compute_quantile_diff_rf(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20, quantile=95):
        y_quantiles_diff = cyext_acv.compute_quantile_diff_rf(x, y, data, y_data, S, self.features, self.thresholds,
                                                              self.children_left,
                                                              self.children_right, self.max_depth, min_node_size,
                                                              classifier, t, quantile)
        return y_quantiles_diff


class ACVTreeAgnostic(BaseTree):

    @staticmethod
    def compute_msdp_clf(X, S, data, model=None, N=10000):
        """
        Compute marginal SDP
        """
        return msdp(X, S, model, data)

    @staticmethod
    def importance_msdp_clf_search(X, data, model=None, C=[[]], minimal=1, pi_level=0.9, r_search_space=None,
                                   stop=True):
        """
        Compute marginal S^\star of model
        """

        return importance_msdp_clf_search(X=X, rg_data=data, model=model, C=C, minimal=minimal,
                                          pi_level=pi_level, r_search_space=r_search_space, stop=stop)

    @staticmethod
    def compute_msdp_reg(X, S, data, model=None, threshold=0.2):
        """
        Compute marginal SDP of regression model
        """
        return msdp_reg(X, S, model, data, threshold)

    @staticmethod
    def importance_msdp_reg_search(X, data, model=None, C=[[]], minimal=1, pi_level=0.9, threshold=0.2,
                                   r_search_space=None, stop=True):
        """
        Compute marginal S^\star of regression model
        """
        return importance_msdp_reg_search(X, rg_data=data, model=model, C=C, minimal=minimal, pi_level=pi_level,
                                          threshold=threshold, r_search_space=r_search_space, stop=stop)

    def compute_cdf_rf(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20):
        sdp = cyext_acv.compute_cdf_rf(x, y, data, y_data, S, self.features, self.thresholds, self.children_left,
                                       self.children_right, self.max_depth, min_node_size, classifier, t)
        return sdp

    def compute_sdp_rf(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20):
        sdp = cyext_acv.compute_sdp_rf(x, y, data, y_data, S, self.features, self.thresholds, self.children_left,
                                       self.children_right, self.max_depth, min_node_size, classifier, t)
        return sdp

    def compute_sdp_rf_v0(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20):
        sdp = cyext_acv.compute_sdp_rf_v0(x, y, data, y_data, S, self.features, self.thresholds, self.children_left,
                                          self.children_right, self.max_depth, min_node_size, classifier, t)
        return sdp

    def compute_sdp_rule(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20):
        sdp, rules = cyext_acv.compute_sdp_rule(x, y, data, y_data, S, self.features, self.thresholds,
                                                self.children_left,
                                                self.children_right, self.max_depth, min_node_size, classifier, t)
        return sdp, rules

    def compute_sdp_rule_v0(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20):
        sdp, rules = cyext_acv.compute_sdp_rule_v0(x, y, data, y_data, S, self.features, self.thresholds,
                                                   self.children_left,
                                                   self.children_right, self.max_depth, min_node_size, classifier, t)
        return sdp, rules

    def compute_sdp_rule_biased(self, x, y, data, S, min_node_size=5, classifier=1, t=20):
        sdp, rules = cyext_acv.compute_sdp_rule_biased(x, y, data, S, self.features, self.thresholds,
                                                       self.children_left,
                                                       self.children_right, self.max_depth, min_node_size, classifier,
                                                       t)
        return sdp, rules

    def compute_sdp_maxrules(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20, pi=0.95):
        sdp, rules, sdp_all, rules_data, w = cyext_acv.compute_sdp_maxrule(x, y, data, y_data, S, self.features,
                                                                           self.thresholds, self.children_left,
                                                                           self.children_right, self.max_depth,
                                                                           min_node_size, classifier, t, pi)
        extend_partition(rules, rules_data, sdp_all, pi=pi, S=S)
        return sdp, rules, sdp_all, rules_data, w

    def compute_sdp_maxrules_biased(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20, pi=0.95):
        sdp, rules, sdp_all, rules_data = cyext_acv.compute_sdp_maxrule_biased(x, y, data, y_data, S, self.features,
                                                                               self.thresholds, self.children_left,
                                                                               self.children_right, self.max_depth,
                                                                               min_node_size, classifier, t, pi)
        extend_partition(rules, rules_data, sdp_all, pi=pi, S=S)
        return sdp, rules, sdp_all, rules_data

    def compute_sdp_maxrules_opti(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20, pi=0.95):
        sdp, rules, sdp_all, rules_data, w = cyext_acv.compute_sdp_maxrule_opti(x, y, data, y_data, S, self.features,
                                                                                self.thresholds, self.children_left,
                                                                                self.children_right, self.max_depth,
                                                                                min_node_size, classifier, t, pi)
        extend_partition(rules, rules_data, sdp_all, pi=pi, S=S)
        return sdp, rules, sdp_all, rules_data, w

    def compute_sdp_maxrules_v0(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20, pi=0.95):
        sdp, rules, sdp_all, rules_data = cyext_acv.compute_sdp_maxrule_v0(x, y, data, y_data, S, self.features,
                                                                           self.thresholds, self.children_left,
                                                                           self.children_right, self.max_depth,
                                                                           min_node_size, classifier, t, pi)
        extend_partition(rules, rules_data, sdp_all, pi=pi, S=S)
        return sdp, rules, sdp_all, rules_data

    def compute_sdp_maxrules_biased_v2(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20, pi=0.95):
        sdp, rules, sdp_all, rules_data, w = cyext_acv.compute_sdp_maxrule_biased_v2(x, y, data, y_data, S,
                                                                                     self.features, self.thresholds,
                                                                                     self.children_left,
                                                                                     self.children_right,
                                                                                     self.max_depth, min_node_size,
                                                                                     classifier, t, pi)
        extend_partition(rules, rules_data, sdp_all, pi=pi, S=S)
        return sdp, rules, sdp_all, rules_data, w

    def importance_sdp_rf(self, x, y, data, y_data, min_node_size=5, classifier=1, t=20,
                          C=[[]], pi_level=0.9, minimal=1, stop=True):

        if x.shape[1] > 10:

            flat_list = [item for t in self.node_idx_trees for sublist in t for item in sublist]
            node_idx = pd.Series(flat_list)
            search_space = []
            for v in (node_idx.value_counts().keys()):
                search_space += [v]
        else:
            search_space = [i for i in range(x.shape[1])]

        sdp = cyext_acv.global_sdp_rf(x, y, data, y_data, self.features, self.thresholds, self.children_left,
                                      self.children_right, self.max_depth, min_node_size, classifier, t, C,
                                      pi_level, minimal, stop, search_space[:10])
        return sdp

    def importance_sdp_rf_v0(self, x, y, data, y_data, min_node_size=5, classifier=1, t=20,
                             C=[[]], pi_level=0.9, minimal=1, stop=True):

        if x.shape[1] > 10:

            flat_list = [item for t in self.node_idx_trees for sublist in t for item in sublist]
            node_idx = pd.Series(flat_list)
            search_space = []
            for v in (node_idx.value_counts().keys()):
                search_space += [v]
        else:
            search_space = [i for i in range(x.shape[1])]

        sdp = cyext_acv.global_sdp_rf_v0(x, y, data, y_data, self.features, self.thresholds, self.children_left,
                                         self.children_right, self.max_depth, min_node_size, classifier, t, C,
                                         pi_level, minimal, stop, search_space[:10])
        return sdp

    def compute_exp_rf(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20):
        exp = cyext_acv.compute_exp_rf(x, y, data, y_data, S, self.features, self.thresholds, self.children_left,
                                       self.children_right, self.max_depth, min_node_size, classifier, t)
        return exp

    def compute_quantile_rf(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20, quantile=95):
        y_quantiles = cyext_acv.compute_quantile_rf(x, y, data, y_data, S, self.features, self.thresholds,
                                                    self.children_left,
                                                    self.children_right, self.max_depth, min_node_size, classifier, t,
                                                    quantile)
        return y_quantiles

    def compute_quantile_diff_rf(self, x, y, data, y_data, S, min_node_size=5, classifier=1, t=20, quantile=95):
        y_quantiles_diff = cyext_acv.compute_quantile_diff_rf(x, y, data, y_data, S, self.features, self.thresholds,
                                                              self.children_left,
                                                              self.children_right, self.max_depth, min_node_size,
                                                              classifier, t, quantile)
        return y_quantiles_diff

    def sufficient_expl_rf(self, x, y, data, y_data, min_node_size=5, classifier=1, t=20,
                           C=[[]], pi_level=0.9, minimal=1, stop=True):

        if x.shape[1] > 10:

            flat_list = [item for t in self.node_idx_trees for sublist in t for item in sublist]
            node_idx = pd.Series(flat_list)
            search_space = []
            for v in (node_idx.value_counts().keys()):
                search_space += [v]
        else:
            search_space = [i for i in range(x.shape[1])]

        sdp = cyext_acv.sufficient_expl_rf(x, y, data, y_data, self.features, self.thresholds, self.children_left,
                                           self.children_right, self.max_depth, min_node_size, classifier, t, C,
                                           pi_level, minimal, stop, search_space[:10])
        return sdp

    def compute_local_sdp(d, sufficient_coal):
        flat = [item for sublist in sufficient_coal for item in sublist]
        flat = pd.Series(flat)
        flat = dict(flat.value_counts() / len(sufficient_coal))
        local_sdp = np.zeros(d)
        for key in flat.keys():
            local_sdp[key] = flat[key]
        return local_sdp

    def compute_exp_shaff(self, X, data, y_data, S, min_node_size=5):
        exp = np.zeros((X.shape[0], 1))
        for i in range(X.shape[0]):
            exp[i] = compute_shaff_exp(X=X[i], S=S, model=self, data=data, Y=y_data, min_node_size=min_node_size)
        return exp

    def compute_sdp_clf_shaff(self, X, y_X, data, y_data, S, min_node_size=5):
        sdp = np.zeros((X.shape[0], 1))
        for i in range(X.shape[0]):
            sdp[i] = compute_shaff_sdp_clf(X=X[i], y_X=y_X[i], S=S, model=self, data=data, Y=y_data,
                                           min_node_size=min_node_size)
        return sdp

    def compute_sdp_shaff(self, X, y_X, t, data, y_data, S, min_node_size=5):
        sdp = np.zeros((X.shape[0], 1))
        for i in range(X.shape[0]):
            sdp[i] = compute_shaff_sdp(X=X[i], t=t, y_X=y_X[i], S=S, model=self, data=data, Y=y_data,
                                       min_node_size=min_node_size)
        return sdp

    def compute_quantile_shaff(self, X, data, y_data, S, min_node_size=5, quantile=95):
        exp = np.zeros((X.shape[0], 1))
        for i in range(X.shape[0]):
            exp[i] = compute_shaff_quantile(X=X[i], S=S, model=self, data=data, Y=y_data, min_node_size=min_node_size,
                                            quantile=quantile)
        return exp



from acv_app.colors import _colors as colors
import matplotlib.pyplot as pl
import matplotlib
import numpy as np
import re
import sys


labels = {
    'MAIN_EFFECT': "SHAP main effect value for\n%s",
    'INTERACTION_VALUE': "SHAP interaction value",
    'INTERACTION_EFFECT': "SHAP interaction value for\n%s and %s",
    'VALUE': "SHAP value (impact on model output)",
    'GLOBAL_VALUE': "mean(|SHAP value|) (average impact on model output magnitude)",
    'VALUE_FOR': "SHAP value for\n%s",
    'PLOT_FOR': "SHAP plot for %s",
    'FEATURE': "Feature %s",
    'FEATURE_VALUE': "Feature value",
    'FEATURE_VALUE_LOW': "Low",
    'FEATURE_VALUE_HIGH': "High",
    'JOINT_VALUE': "Joint SHAP value",
    'MODEL_OUTPUT': "Model output value"
}

def format_value(s, format_str):
    """ Strips trailing zeros and uses a unicode minus sign.
    """

    if not issubclass(type(s), str):
        s = format_str % s
    s = re.sub(r'\.?0+$', '', s)
    if s[0] == "-":
        s = u"\u2212" + s[1:]
    return s


def safe_isinstance(obj, class_path_str):
    """
    Acts as a safe version of isinstance without having to explicitly
    import packages which may not exist in the users environment.
    Checks if obj is an instance of type specified by class_path_str.
    Parameters
    ----------
    obj: Any
        Some object you want to test against
    class_path_str: str or list
        A string or list of strings specifying full class paths
        Example: `sklearn.ensemble.RandomForestRegressor`
    Returns
    --------
    bool: True if isinstance is true and the package exists, False otherwise
    """
    if isinstance(class_path_str, str):
        class_path_strs = [class_path_str]
    elif isinstance(class_path_str, list) or isinstance(class_path_str, tuple):
        class_path_strs = class_path_str
    else:
        class_path_strs = ['']

    # try each module path in order
    for class_path_str in class_path_strs:
        if "." not in class_path_str:
            raise ValueError("class_path_str must be a string or list of strings specifying a full \
                module path to a class. Eg, 'sklearn.ensemble.RandomForestRegressor'")

        # Splits on last occurence of "."
        module_name, class_name = class_path_str.rsplit(".", 1)

        # here we don't check further if the model is not imported, since we shouldn't have
        # an object of that types passed to us if the model the type is from has never been
        # imported. (and we don't want to import lots of new modules for no reason)
        if module_name not in sys.modules:
            continue

        module = sys.modules[module_name]

        # Get class
        _class = getattr(module, class_name, None)

        if _class is None:
            continue

        if isinstance(obj, _class):
            return True

    return False

def waterfall(base_values, values, features, feature_names, max_display=10, show=True):
    """ Plots an explantion of a single prediction as a waterfall plot.
    The SHAP value of a feature represents the impact of the evidence provided by that feature on the model's
    output. The waterfall plot is designed to visually display how the SHAP values (evidence) of each feature
    move the model output from our prior expectation under the background data distribution, to the final model
    prediction given the evidence of all the features. Features are sorted by the magnitude of their SHAP values
    with the smallest magnitude features grouped together at the bottom of the plot when the number of features
    in the models exceeds the max_display parameter.

    Parameters
    ----------
    shap_values : Explanation
        A one-dimensional Explanation object that contains the feature values and SHAP values to plot.
    max_display : str
        The maximum number of features to plot.
    show : bool
        Whether matplotlib.pyplot.show() is called before returning. Setting this to False allows the plot
        to be customized further after it has been created.
    """

    # # Turn off interactive plot
    # if show is False:
    #     plt.ioff()

    fig = pl.figure()
    # base_values = shap_values.base_values

    # features = shap_values.data
    # feature_names = shap_values.feature_names
    lower_bounds = None
    upper_bounds = None
    # values = shap_values.values

    # make sure we only have a single output to explain
    if (type(base_values) == np.ndarray and len(base_values) > 0) or type(base_values) == list:
        raise Exception("waterfall_plot requires a scalar base_values of the model output as the first " \
                        "parameter, but you have passed an array as the first parameter! " \
                        "Try shap.waterfall_plot(explainer.base_values[0], values[0], X[0]) or " \
                        "for multi-output models try " \
                        "shap.waterfall_plot(explainer.base_values[0], values[0][0], X[0]).")

    # make sure we only have a single explanation to plot
    if len(values.shape) == 2:
        raise Exception(
            "The waterfall_plot can currently only plot a single explanation but a matrix of explanations was passed!")

    # unwrap pandas series
    if safe_isinstance(features, "pandas.core.series.Series"):
        if feature_names is None:
            feature_names = list(features.index)
        features = features.values

    # fallback feature names
    if feature_names is None:
        feature_names = np.array([labels['FEATURE'] % str(i) for i in range(len(values))])

    # init variables we use for tracking the plot locations
    num_features = min(max_display, len(values))
    row_height = 0.5
    rng = range(num_features - 1, -1, -1)
    order = np.argsort(-np.abs(values))
    pos_lefts = []
    pos_inds = []
    pos_widths = []
    pos_low = []
    pos_high = []
    neg_lefts = []
    neg_inds = []
    neg_widths = []
    neg_low = []
    neg_high = []
    loc = base_values + values.sum()
    yticklabels = ["" for i in range(num_features + 1)]

    # size the plot based on how many features we are plotting
    pl.gcf().set_size_inches(8, num_features * row_height + 1.5)

    # see how many individual (vs. grouped at the end) features we are plotting
    if num_features == len(values):
        num_individual = num_features
    else:
        num_individual = num_features - 1

    # compute the locations of the individual features and plot the dashed connecting lines
    for i in range(num_individual):
        sval = values[order[i]]
        loc -= sval
        if sval >= 0:
            pos_inds.append(rng[i])
            pos_widths.append(sval)
            if lower_bounds is not None:
                pos_low.append(lower_bounds[order[i]])
                pos_high.append(upper_bounds[order[i]])
            pos_lefts.append(loc)
        else:
            neg_inds.append(rng[i])
            neg_widths.append(sval)
            if lower_bounds is not None:
                neg_low.append(lower_bounds[order[i]])
                neg_high.append(upper_bounds[order[i]])
            neg_lefts.append(loc)
        if num_individual != num_features or i + 4 < num_individual:
            pl.plot([loc, loc], [rng[i] - 1 - 0.4, rng[i] + 0.4], color="#bbbbbb", linestyle="--", linewidth=0.5,
                    zorder=-1)
        if features is None:
            yticklabels[rng[i]] = feature_names[order[i]]
        else:
            yticklabels[rng[i]] = format_value(features[order[i]], "%0.03f") + " = " + feature_names[order[i]]

            # add a last grouped feature to represent the impact of all the features we didn't show
    if num_features < len(values):
        yticklabels[0] = "%d other features" % (len(values) - num_features + 1)
        remaining_impact = base_values - loc
        if remaining_impact < 0:
            pos_inds.append(0)
            pos_widths.append(-remaining_impact)
            pos_lefts.append(loc + remaining_impact)
            c = colors.red_rgb
        else:
            neg_inds.append(0)
            neg_widths.append(-remaining_impact)
            neg_lefts.append(loc + remaining_impact)
            c = colors.blue_rgb

    points = pos_lefts + list(np.array(pos_lefts) + np.array(pos_widths)) + neg_lefts + list(
        np.array(neg_lefts) + np.array(neg_widths))
    dataw = np.max(points) - np.min(points)

    # draw invisible bars just for sizing the axes
    label_padding = np.array([0.1 * dataw if w < 1 else 0 for w in pos_widths])
    pl.barh(pos_inds, np.array(pos_widths) + label_padding + 0.02 * dataw, left=np.array(pos_lefts) - 0.01 * dataw,
            color=colors.red_rgb, alpha=0)
    label_padding = np.array([-0.1 * dataw if -w < 1 else 0 for w in neg_widths])
    pl.barh(neg_inds, np.array(neg_widths) + label_padding - 0.02 * dataw, left=np.array(neg_lefts) + 0.01 * dataw,
            color=colors.blue_rgb, alpha=0)

    # define variable we need for plotting the arrows
    head_length = 0.08
    bar_width = 0.8
    xlen = pl.xlim()[1] - pl.xlim()[0]
    fig = pl.gcf()
    ax = pl.gca()
    xticks = ax.get_xticks()
    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    width, height = bbox.width, bbox.height
    bbox_to_xscale = xlen / width
    hl_scaled = bbox_to_xscale * head_length
    renderer = fig.canvas.get_renderer()

    # draw the positive arrows
    for i in range(len(pos_inds)):
        dist = pos_widths[i]
        arrow_obj = pl.arrow(
            pos_lefts[i], pos_inds[i], max(dist - hl_scaled, 0.000001), 0,
            head_length=min(dist, hl_scaled),
            color=colors.red_rgb, width=bar_width,
            head_width=bar_width
        )

        if pos_low is not None and i < len(pos_low):
            pl.errorbar(
                pos_lefts[i] + pos_widths[i], pos_inds[i],
                xerr=np.array([[pos_widths[i] - pos_low[i]], [pos_high[i] - pos_widths[i]]]),
                ecolor=colors.light_red_rgb
            )

        txt_obj = pl.text(
            pos_lefts[i] + 0.5 * dist, pos_inds[i], format_value(pos_widths[i], '%+0.02f'),
            horizontalalignment='center', verticalalignment='center', color="white",
            fontsize=12
        )
        text_bbox = txt_obj.get_window_extent(renderer=renderer)
        arrow_bbox = arrow_obj.get_window_extent(renderer=renderer)

        # if the text overflows the arrow then draw it after the arrow
        if text_bbox.width > arrow_bbox.width:
            txt_obj.remove()

            txt_obj = pl.text(
                pos_lefts[i] + (5 / 72) * bbox_to_xscale + dist, pos_inds[i], format_value(pos_widths[i], '%+0.02f'),
                horizontalalignment='left', verticalalignment='center', color=colors.red_rgb,
                fontsize=12
            )

    # draw the negative arrows
    for i in range(len(neg_inds)):
        dist = neg_widths[i]

        arrow_obj = pl.arrow(
            neg_lefts[i], neg_inds[i], -max(-dist - hl_scaled, 0.000001), 0,
            head_length=min(-dist, hl_scaled),
            color=colors.blue_rgb, width=bar_width,
            head_width=bar_width
        )

        if neg_low is not None and i < len(neg_low):
            pl.errorbar(
                neg_lefts[i] + neg_widths[i], neg_inds[i],
                xerr=np.array([[neg_widths[i] - neg_low[i]], [neg_high[i] - neg_widths[i]]]),
                ecolor=colors.light_blue_rgb
            )

        txt_obj = pl.text(
            neg_lefts[i] + 0.5 * dist, neg_inds[i], format_value(neg_widths[i], '%+0.02f'),
            horizontalalignment='center', verticalalignment='center', color="white",
            fontsize=12
        )
        text_bbox = txt_obj.get_window_extent(renderer=renderer)
        arrow_bbox = arrow_obj.get_window_extent(renderer=renderer)

        # if the text overflows the arrow then draw it after the arrow
        if text_bbox.width > arrow_bbox.width:
            txt_obj.remove()

            txt_obj = pl.text(
                neg_lefts[i] - (5 / 72) * bbox_to_xscale + dist, neg_inds[i], format_value(neg_widths[i], '%+0.02f'),
                horizontalalignment='right', verticalalignment='center', color=colors.blue_rgb,
                fontsize=12
            )

    # draw the y-ticks twice, once in gray and then again with just the feature names in black
    ytick_pos = list(range(num_features)) + list(
        np.arange(num_features) + 1e-8)  # The 1e-8 is so matplotlib 3.3 doesn't try and collapse the ticks
    pl.yticks(ytick_pos, yticklabels[:-1] + [l.split('=')[-1] for l in yticklabels[:-1]], fontsize=13)

    # put horizontal lines for each feature row
    for i in range(num_features):
        pl.axhline(i, color="#cccccc", lw=0.5, dashes=(1, 5), zorder=-1)

    # mark the prior expected value and the model prediction
    pl.axvline(base_values, 0, 1 / num_features, color="#bbbbbb", linestyle="--", linewidth=0.5, zorder=-1)
    fx = base_values + values.sum()
    pl.axvline(fx, 0, 1, color="#bbbbbb", linestyle="--", linewidth=0.5, zorder=-1)

    # clean up the main axis
    pl.gca().xaxis.set_ticks_position('bottom')
    pl.gca().yaxis.set_ticks_position('none')
    pl.gca().spines['right'].set_visible(False)
    pl.gca().spines['top'].set_visible(False)
    pl.gca().spines['left'].set_visible(False)
    ax.tick_params(labelsize=13)
    # pl.xlabel("\nModel output", fontsize=12)

    # draw the E[f(X)] tick mark
    xmin, xmax = ax.get_xlim()
    ax2 = ax.twiny()
    ax2.set_xlim(xmin, xmax)
    ax2.set_xticks(
        [base_values, base_values + 1e-8])  # The 1e-8 is so matplotlib 3.3 doesn't try and collapse the ticks
    ax2.set_xticklabels(["\n$E[f(X)]$", "\n$ = " + format_value(base_values, "%0.03f") + "$"], fontsize=12, ha="left")
    ax2.spines['right'].set_visible(False)
    ax2.spines['top'].set_visible(False)
    ax2.spines['left'].set_visible(False)

    # draw the f(x) tick mark
    ax3 = ax2.twiny()
    ax3.set_xlim(xmin, xmax)
    ax3.set_xticks([base_values + values.sum(),
                    base_values + values.sum() + 1e-8])  # The 1e-8 is so matplotlib 3.3 doesn't try and collapse the ticks
    ax3.set_xticklabels(["$f(x)$", "$ = " + format_value(fx, "%0.03f") + "$"], fontsize=12, ha="left")
    tick_labels = ax3.xaxis.get_majorticklabels()
    tick_labels[0].set_transform(
        tick_labels[0].get_transform() + matplotlib.transforms.ScaledTranslation(-10 / 72., 0, fig.dpi_scale_trans))
    tick_labels[1].set_transform(
        tick_labels[1].get_transform() + matplotlib.transforms.ScaledTranslation(12 / 72., 0, fig.dpi_scale_trans))
    tick_labels[1].set_color("#999999")
    ax3.spines['right'].set_visible(False)
    ax3.spines['top'].set_visible(False)
    ax3.spines['left'].set_visible(False)

    # adjust the position of the E[f(X)] = x.xx label
    tick_labels = ax2.xaxis.get_majorticklabels()
    tick_labels[0].set_transform(
        tick_labels[0].get_transform() + matplotlib.transforms.ScaledTranslation(-20 / 72., 0, fig.dpi_scale_trans))
    tick_labels[1].set_transform(
        tick_labels[1].get_transform() + matplotlib.transforms.ScaledTranslation(22 / 72., -1 / 72.,
                                                                                 fig.dpi_scale_trans))

    tick_labels[1].set_color("#999999")

    # color the y tick labels that have the feature values as gray
    # (these fall behind the black ones with just the feature name)
    tick_labels = ax.yaxis.get_majorticklabels()
    for i in range(num_features):
        tick_labels[i].set_color("#999999")

    return fig


def bar_legacy(shap_values, features=None, feature_names=None, max_display=None, show=True):
    # unwrap pandas series
    if str(type(features)) == "<class 'pandas.core.series.Series'>":
        if feature_names is None:
            feature_names = list(features.index)
        features = features.values

    if feature_names is None:
        feature_names = np.array([labels['FEATURE'] % str(i) for i in range(len(shap_values))])

    if max_display is None:
        max_display = 7
    else:
        max_display = min(len(feature_names), max_display)

    feature_order = np.argsort(-np.abs(shap_values))

    #
    feature_inds = feature_order[:max_display]
    y_pos = np.arange(len(feature_inds), 0, -1)
    pl.barh(
        y_pos, shap_values[feature_inds],
        0.7, align='center',
        color=[colors.red_rgb if shap_values[feature_inds[i]] > 0 else colors.blue_rgb for i in range(len(y_pos))]
    )
    for y in range(len(y_pos)):
        pl.text(shap_values[feature_inds][y]+0.1, y_pos[y],  round(shap_values[feature_inds][y], 3))

    pl.yticks(y_pos, fontsize=13)
    if features is not None:
        features = list(features)

        # try and round off any trailing zeros after the decimal point in the feature values
        for i in range(len(features)):
            try:
                if round(features[i]) == features[i]:
                    features[i] = int(features[i])
            except TypeError:
                pass  # features[i] must not be a number
    yticklabels = []
    for i in feature_inds:
        if features is not None:
            yticklabels.append(feature_names[i] + " = " + str(features[i]))
        else:
            yticklabels.append(feature_names[i])
    pl.gca().set_yticklabels(yticklabels)
    pl.gca().xaxis.set_ticks_position('bottom')
    pl.gca().yaxis.set_ticks_position('none')
    pl.gca().spines['right'].set_visible(False)
    pl.gca().spines['top'].set_visible(False)
    # pl.gca().spines['left'].set_visible(False)

    pl.xlabel("SHAP value (impact on model output)")

    return fig


from abc import abstractmethod
import scipy.special
import json
import struct
import cext_acv
import warnings
from .py_acv import *
from .utils import rebuild_tree
from distutils.version import LooseVersion


# This is based on https://github.com/slundberg/shap/blob/master/shap/explainers/_tree.py
class BaseAgnosTree:
    """ An ensemble of decision trees.

    This object provides a common interface to many different types of models.
    """

    def __init__(self, model, data_dim=None, data=None, data_missing=None, cache=False, cache_normalized=False, multi_threads=True,
                 C=[[]]):
        self.model_type = "internal"
        self.trees = None
        self.base_offset = 0
        self.model_output = None
        self.objective = None  # what we explain when explaining the loss of the model
        self.tree_output = None  # what are the units of the values in the leaves of the trees
        self.internal_dtype = np.float64
        self.input_dtype = np.float64  # for sklearn we need to use np.float32 to always get exact matches to their predictions
        self.data = data
        self.data_missing = data_missing
        self.fully_defined_weighting = True  # does the background dataset land in every leaf (making it valid for the tree_path_dependent method)
        self.tree_limit = None  # used for limiting the number of trees we use by default (like from early stopping)
        self.num_stacked_models = 1  # If this is greater than 1 it means we have multiple stacked models with the same number of trees in each model (XGBoost multi-output style)
        self.cat_feature_indices = None  # If this is set it tells us which features are treated categorically
        self.model = model
        self.cache = cache
        self.cache_normalized = cache_normalized
        self.C = C
        self.multi_threads = multi_threads
        self.data_dim = data_dim

        if safe_isinstance(model, ["skranger.ensemble.classifier.RangerForestClassifier"]):
            assert hasattr(model, "estimators_"), "Model has no `estimators_`! Have you called `model.fit`?"
            self.internal_dtype = model.estimators_[0].tree_.value.dtype.type
            self.input_dtype = np.float32
            scaling = 1.0 / len(model.estimators_)  # output is average of trees
            # self.scaling = scaling
            self.trees = [SingleTree(e.tree_, normalize=False, scaling=scaling, data=data, data_missing=data_missing) for
                          e in model.estimators_]
            self.tree_output = "probability"

        elif safe_isinstance(model, ["skranger.ensemble.regressor.RangerForestRegressor"]):
            assert hasattr(model, "estimators_"), "Model has no `estimators_`! Have you called `model.fit`?"
            self.internal_dtype = model.estimators_[0].tree_.value.dtype.type
            self.input_dtype = np.float32
            scaling = 1.0 / len(model.estimators_)  # output is average of trees
            # self.scaling = scaling
            self.trees = [SingleTree(e.tree_, scaling=scaling, data=data, data_missing=data_missing) for e in
                          model.estimators_]
            self.tree_output = "raw_value"

        else:
            raise Exception("Model type not yet supported by TreeExplainer: " + str(type(model)))

        # build a dense numpy version of all the tree objects
        if self.trees is not None and self.trees:
            max_nodes = np.max([len(t.values) for t in self.trees])
            assert len(np.unique([t.values.shape[1] for t in
                                  self.trees])) == 1, "All trees in the ensemble must have the same output dimension!"
            num_trees = len(self.trees)
            if self.num_stacked_models > 1:
                assert len(
                    self.trees) % self.num_stacked_models == 0, "Only stacked models with equal numbers of trees are supported!"
                assert self.trees[0].values.shape[
                           1] == 1, "Only stacked models with single outputs per model are supported!"
                self.num_outputs = self.num_stacked_models
            else:
                self.num_outputs = self.trees[0].values.shape[1]

            # important to be -1 in unused sections!! This way we can tell which entries are valid.
            self.children_left = -np.ones((num_trees, max_nodes), dtype=np.int32)
            self.children_right = -np.ones((num_trees, max_nodes), dtype=np.int32)
            self.children_default = -np.ones((num_trees, max_nodes), dtype=np.int32)
            self.features = -np.ones((num_trees, max_nodes), dtype=np.int32)

            self.thresholds = np.zeros((num_trees, max_nodes), dtype=self.internal_dtype)
            self.values = np.zeros((num_trees, max_nodes, self.num_outputs), dtype=self.internal_dtype)
            self.node_sample_weight = np.zeros((num_trees, max_nodes), dtype=self.internal_dtype)

            self.partition_leaves_trees = []
            self.node_idx_trees = []
            # self.data_leaves_trees = []
            self.leaf_idx_trees = []
            self.leaves_nb = []
            self.scalings = []
            for i in tqdm(range(num_trees)):
                self.scalings.append(self.trees[i].scaling)
                self.children_left[i, :len(self.trees[i].children_left)] = self.trees[i].children_left
                self.children_right[i, :len(self.trees[i].children_right)] = self.trees[i].children_right
                self.children_default[i, :len(self.trees[i].children_default)] = self.trees[i].children_default
                self.features[i, :len(self.trees[i].features)] = self.trees[i].features
                self.thresholds[i, :len(self.trees[i].thresholds)] = self.trees[i].thresholds
                if self.num_stacked_models > 1:
                    # stack_pos = int(i // (num_trees / self.num_stacked_models))
                    stack_pos = i % self.num_stacked_models
                    self.values[i, :len(self.trees[i].values[:, 0]), stack_pos] = self.trees[i].values[:, 0]
                else:
                    self.values[i, :len(self.trees[i].values)] = self.trees[i].values
                self.node_sample_weight[i, :len(self.trees[i].node_sample_weight)] = self.trees[i].node_sample_weight

                # ensure that the passed background dataset lands in every leaf
                if np.min(self.trees[i].node_sample_weight) <= 0:
                    self.fully_defined_weighting = False

                self.leaf_idx = [idx for idx in range(len(self.trees[i].features))
                                 if self.trees[i].children_left[idx] < 0]

                self.leaves_nb.append(len(self.leaf_idx))

                self.partition_leaves = []
                self.node_idx = []
                self.max_var = []
                # self.data_leaves = []
                for leaf_id in self.leaf_idx:
                    node_id = [-1]
                    partition_leaf = [np.array([[-np.inf, np.inf]]) for idx2 in range(self.data_dim)]
                    _ = get_partition(leaf_id, partition_leaf, node_id, self.trees[i].children_left,
                                      self.trees[i].children_right, self.trees[i].features, self.trees[i].thresholds)

                    self.partition_leaves.append(np.squeeze(np.array(partition_leaf)))
                    self.node_idx.append(list(set(node_id[1:])))
                    self.max_var.append(len(self.node_idx[-1]))
                    # self.data_leaves.append(np.array([(self.data[:, s] <= self.partition_leaves[-1][s, 1]) * \
                    #                                       (self.data[:, s] > self.partition_leaves[-1][s, 0])
                    #                                       for s in range(self.data.shape[1])], dtype=np.int).transpose())

                self.partition_leaves_trees.append(self.partition_leaves)
                # self.data_leaves_trees.append(self.data_leaves)

                self.node_idx_trees.append(self.node_idx)
                self.leaf_idx_trees.append(self.leaf_idx)

            leaf_idx_trees = -np.ones(shape=(len(self.leaves_nb), np.max(self.leaves_nb)), dtype=np.int)
            partition_leaves_trees = -np.ones(
                shape=(len(self.leaves_nb), np.max(self.leaves_nb), self.data_dim, 2))
            # data_leaves_trees = -np.ones(shape=(len(self.leaves_nb), np.max(self.leaves_nb), self.data.shape[0], self.data.shape[1]), dtype=np.int)
            for i in range(len(self.leaves_nb)):
                leaf_idx_trees[i, :self.leaves_nb[i]] = np.array(self.leaf_idx_trees[i], dtype=np.int)
                partition_leaves_trees[i, :self.leaves_nb[i]] = np.array(self.partition_leaves_trees[i])
                # data_leaves_trees[i, :self.leaves_nb[i]] = np.array(self.data_leaves_trees[i], dtype=np.int)

            self.leaf_idx_trees = leaf_idx_trees
            self.partition_leaves_trees = partition_leaves_trees
            self.leaves_nb = np.array(self.leaves_nb, dtype=np.int)
            self.scalings = np.array(self.scalings, dtype=np.float)
            # self.data = np.array(self.data, dtype=np.float)
            self.max_var = np.max(self.max_var)
            # self.data_leaves_trees = data_leaves_trees

            # if safe_isinstance(model, ["xgboost.sklearn.XGBClassifier",
            #                            "catboost.core.CatBoostClassifier", "lightgbm.sklearn.LGBMClassifier"]) and \
            #         self.num_outputs == 1:
            #     p = np.exp(self.values)/(1 + np.exp(self.values))
            #     print(np.max(p), np.min(1-p))
            #     self.values = np.concatenate([1-p, p], axis=2)
            #     self.num_outputs = 2

            self.num_nodes = np.array([len(t.values) for t in self.trees], dtype=np.int32)
            self.max_depth = np.max([t.max_depth for t in self.trees])

class SingleTree:
    """ A single decision tree.

    The primary point of this object is to parse many different tree types into a common format.
    """

    def __init__(self, tree, normalize=False, scaling=1.0, data=None, data_missing=None):
        self.scaling = scaling
        self.children_left = tree.children_left.astype(np.int32)
        self.children_right = tree.children_right.astype(np.int32)
        self.children_default = self.children_left  # missing values not supported in sklearn
        self.features = tree.feature.astype(np.int32)
        self.thresholds = tree.threshold.astype(np.float64)
        self.values = tree.value.reshape(tree.value.shape[0], tree.value.shape[1] * tree.value.shape[2])
        if normalize:
            self.values = (self.values.T / self.values.sum(1)).T
        self.values = self.values * scaling
        self.node_sample_weight = tree.weighted_n_node_samples.astype(np.float64)

        self.max_depth = cext_acv.compute_expectations(
            self.children_left, self.children_right, self.node_sample_weight,
            self.values
        )


import scipy.stats as st
import sys
import itertools
from collections import defaultdict
from operator import itemgetter
import string
import random
import scipy.stats as stats
from typing import Tuple, List
from sklearn.cluster import AffinityPropagation
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.preprocessing import OrdinalEncoder


def plot_feature_importance(importance, names, model_type, xlabel='SHAP values', title=' '):
    # Create arrays from feature importance and feature names
    feature_importance = np.array(importance)
    feature_names = np.array(names)

    # Create a DataFrame using a Dictionary
    data = {'feature_names': feature_names, 'feature_importance': feature_importance}
    fi_df = pd.DataFrame(data)

    # Sort the DataFrame in order decreasing feature importance
    fi_df.sort_values(by=['feature_importance'], ascending=False, inplace=True)

    # Define size of bar plot
    plt.figure(figsize=(10, 8))
    # Plot Searborn bar chart
    sns.barplot(x=fi_df['feature_importance'], y=fi_df['feature_names'])
    # Add chart labels
    plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel('FEATURE NAMES')


def plot_feature_importance_10(importance, names, model_type, xlabel='SHAP values', title=' '):
    # Create arrays from feature importance and feature names
    feature_importance = np.array(importance)
    feature_names = np.array(names)
    seaborn_colors = sns.color_palette("tab10")
    colors = {names[i]: seaborn_colors[i] for i in range(len(names))}

    # Create a DataFrame using a Dictionary
    data = {'feature_names': feature_names, 'feature_importance': feature_importance}
    fi_df = pd.DataFrame(data)

    # Sort the DataFrame in order decreasing feature importance
    fi_df.sort_values(by=['feature_importance'], ascending=False, inplace=True)

    # Define size of bar plot
    plt.figure(figsize=(10, 8))
    sns.set(font_scale=1.5)
    # Plot Searborn bar chart
    sns.barplot(x=fi_df['feature_importance'], y=fi_df['feature_names'], palette=colors)
    # Add chart labels
    plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel('FEATURE NAMES')


def bar_plot(values_1, values_2, values_3, labels, variables_name, title):
    x = np.arange(len(variables_name))  # the label locations
    width = 0.2  # the width of the bars

    fig, ax = plt.subplots(dpi=250)
    rects1 = ax.bar(x - width / 2, values_1, width, label=labels[0])
    rects2 = ax.bar(x + width / 2, values_2, width, label=labels[1])
    rects3 = ax.bar(x + 1.5 * width, values_3, width, label=labels[2])

    # Add some text for labels, title and custom x-axis tick labels, etc.
    ax.set_ylabel('Shapley values')
    ax.set_title(title)
    ax.set_xticks(x)
    ax.set_xticklabels(variables_name)
    ax.legend()

    def autolabel(rects):
        """Attach a text label above each bar in *rects*, displaying its height."""
        for rect in rects:
            height = rect.get_height()
            ax.annotate('{}'.format(height),
                        xy=(rect.get_x() + rect.get_width() / 2, height),
                        xytext=(0, 3),  # 3 points vertical offset
                        textcoords="offset points",
                        ha='center', va='bottom', color='black')

    autolabel(rects1)
    autolabel(rects2)
    autolabel(rects3)

    fig.tight_layout()

    plt.show()


def compute_fcluster_bydistance(data, dist_type='kendall'):
    d = data.shape[1]
    if dist_type == 'kendall':
        dist_matrix = np.zeros((d, d))
        for i in range(d):
            for j in range(d):
                tau, p_value = stats.kendalltau(data[:, i], data[:, j])
                dist_matrix[i, j] = tau
    elif dist_type == 'corr':
        dist_matrix = np.cov(data.T)
    else:
        raise ValueError("This type of distance is not implemented")

    #     print(dist_matrix)
    clustering = AffinityPropagation(random_state=5)
    coal_idx = clustering.fit_predict(np.abs(dist_matrix))
    coal = []
    for c in np.unique(coal_idx):
        t = np.argwhere(c == coal_idx)
        t = list(t.reshape(-1))
        coal.append(t)
    return coal


def ecart_model(model, X, y):
    y_predict = model.predict(X)
    error = (y_predict - y) ** 2
    print('mse = {} -- max = {} -- min = {} -- q0.95 = {} -- q0.25 {}'.format(np.mean(error),
                                                                              np.max(error),
                                                                              np.min(error),
                                                                              np.quantile(error, 0.95),
                                                                              np.quantile(error, 0.5)))
    return np.mean(error), np.quantile(error, 0.95), np.quantile(error, 0.25)


def get_given_data(x, idx, N):
    val = np.array([x[idx]])
    val = np.tile(val, N)
    return val


# def gen_data_by_cat(x_in, cat_va, prob, N, S, S_bar):
#     # to do: make it real function of S, c_index, etc
#
#     mixture_idx = np.random.choice(cat_va, size=N, replace=True, p=prob)
#
#     rg = {key: sampleMVN(np.sum(mixture_idx == key), self.mean[key],
#                          self.cov[key], S_bar, S, x_in[S]) for key in cat_va
#           if np.sum(mixture_idx == key) != 0}
#
#     rg = np.concatenate([np.concatenate([rg[key], np.tile(self.dummy[key],
#                                                           (np.sum(mixture_idx == key), 1))], axis=1) for key in
#                          rg.keys()])
#
#     rg_data = pd.DataFrame(rg, columns=[str(s) for s in S_bar] + [str(ca) for ca in self.cat_index])
#
#     for val_id in S:
#         rg_data[str(val_id)] = get_given_data(val_id)
#
#     rg_data = rg_data[sorted(rg_data.columns)]
#     return rg_data


def nb_row(row, up, low, S):
    return np.prod([(row[s] <= up[s]) * (row[s] > low[s]) for s in S])


def get_partition_tree(tree, leaf_id, part):
    a = tree.children_left.copy()
    b = tree.children_right.copy()
    f = tree.feature.copy()
    t = tree.threshold.copy()
    r_w = tree.weighted_n_node_samples.copy()
    r = tree.n_node_samples.copy()
    left = np.where(tree.children_left == leaf_id)[0]
    right = np.where(tree.children_right == leaf_id)[0]

    if (len(left) == 0) * (len(right) == 0):
        return (part)

    else:
        if len(right) != 0:
            right = int(right[0])

            part[f[right]] = np.concatenate((part[f[right]], np.array([[t[right], np.inf]])))
            return get_partition_tree(tree, right, part)
        else:
            left = int(left[0])
            part[f[left]] = np.concatenate((part[f[left]], np.array([[-np.inf, t[left]]])))
            return get_partition_tree(tree, left, part)


def get_final_partition(part):
    final_partition = {}
    for i, var_part in enumerate(part):
        final_partition[i] = [np.max(var_part[:, 0]), np.min(var_part[:, 1])]
    return final_partition


def leaves_proba(tree, leaf_id, data, S):
    partition_leaves = [np.array([[-np.inf, np.inf]]) for i in range(data.shape[1])]
    partition_leaves = get_partition_tree(tree, leaf_id, partition_leaves)
    partition_leaves = pd.DataFrame(get_final_partition(partition_leaves))[S]

    low = partition_leaves.iloc[0]
    up = partition_leaves.iloc[1]

    section_x = np.prod([(data[:, s] <= up[s]) * (data[:, s] >= low[s]) for s in S], axis=0)
    return np.sum(section_x)


def l1_norm(x, dim=1):
    return np.sum(np.abs(x), axis=1)


def rebuild_tree(parent_id, tree, data, y):
    tree.weighted_n_node_samples[parent_id] = len(data)
    tree.value[parent_id][0][0] = np.mean(y)

    if tree.children_right[parent_id] < 0:
        return 0
    else:
        right = tree.children_right[parent_id]
        left = tree.children_left[parent_id]

        left_cond = data[:, tree.feature[parent_id]] <= tree.threshold[parent_id]
        data_left = data[left_cond]
        y_left = y[left_cond]

        right_cond = data[:, tree.feature[parent_id]] > tree.threshold[parent_id]
        data_right = data[right_cond]
        y_right = y[right_cond]

        rebuild_tree(left, tree, data_left, y_left)
        rebuild_tree(right, tree, data_right, y_right)


def rebuild_acvtree(parent_id, tree, data, y):
    tree.node_sample_weight[parent_id] = data.shape[0]
    tree.values[parent_id] = np.mean(y, axis=0)

    if tree.children_right[parent_id] < 0:
        return 0
    else:
        right = tree.children_right[parent_id]
        left = tree.children_left[parent_id]

        left_cond = data[:, tree.features[parent_id]] <= tree.thresholds[parent_id]
        data_left = data[left_cond]
        y_left = y[left_cond]

        right_cond = data[:, tree.features[parent_id]] > tree.thresholds[parent_id]
        data_right = data[right_cond]
        y_right = y[right_cond]

        rebuild_acvtree(left, tree, data_left, y_left)
        rebuild_acvtree(right, tree, data_right, y_right)


def condMVN(mean, cov, set_bar, set, x):
    if set == []:
        return mean, cov
    else:
        mean_cond = mean[set_bar] + np.matmul(np.matmul(cov[set_bar][:, set],
                                                        np.linalg.inv(cov[set][:, set])), x - mean[set])

        cov_cond = cov[set_bar][:, set_bar] - np.matmul(
            np.matmul(cov[set_bar][:, set], np.linalg.inv(cov[set][:, set])),
            cov[set][:, set_bar])

        return mean_cond, cov_cond


def marMVN(mean, cov, set_bar, set, x):
    if set == []:
        return mean, cov
    else:
        mean_cond = mean[set_bar]
        cov_cond = cov[set_bar][:, set_bar]
        return mean_cond, cov_cond


def sampleMVN(n, mean, cov, set_bar, set, x):
    mean_cond, cov_cond = condMVN(mean, cov, set_bar, set, x)
    sample = st.multivariate_normal(mean_cond, cov_cond).rvs(n)
    return np.reshape(sample, (n, len(set_bar)))


def sampleMarMVN(n, mean, cov, set_bar, set):
    mean_cond, cov_cond = marMVN(mean, cov, set_bar, set)
    sample = st.multivariate_normal(mean_cond, cov_cond).rvs(n)
    return np.reshape(sample, (n, len(set_bar)))


def chain_l(l):
    chain = []
    if type(l) == tuple:
        for it in l:
            if type(it) != list:
                chain.append(it)
            elif type(it) == list and len(it) > 1:
                chain = chain + it
            else:
                raise ValueError('problem...')
    else:
        chain = l
    return chain


def convert_list(l):
    if type(l) == list:
        return l
    else:
        return [l]


def convert_tuple(l):
    if type(l) == tuple:
        return l
    elif type(l) == list:
        return tuple(l)
    else:
        return (l,)


def linear_regression(coefs, x):
    return np.sum(coefs * x, axis=1)


def linear_regression_0(coefs, x):
    return np.sum(coefs * x)


# utils for classifer data generation


def return_positive_semi_definite_matrix(n_dim: int) -> np.ndarray:
    """Return positive semi-definite matrix.

    Args:
        n_dim (int): size of square matrix to return
    Returns:
        p (np.array): positive semi-definite array of shape (n_dim, n_dim)
    """
    m = np.random.randn(n_dim, n_dim)
    p = np.dot(m, m.T)
    return p


def sigmoid(x: np.array) -> np.array:
    """Return sigmoid(x) for some activations x.

    Args:
        x (np.array): input activations
    Returns:
        s (np.array): sigmoid(x)
    """
    s = 1 / (1 + np.exp(-x))
    return s


def return_weak_features_and_targets(
        num_features: int,
        num_samples: int,
        mixing_factor: float,
) -> Tuple[np.array, np.array, np.array, np.array]:
    """Return weakly predictive features and a target variable.

    Create a multivariate Gaussian-distributed set of features and a
    response variable that is conditioned on a weighted sum of the features.

    Args:
        num_features (int): number of variables in Gaussian distribution
        num_samples (int): number of samples to take
        mixing_factor (float): squashes the weighted sum into the linear
            regime of a sigmoid.  Smaller numbers squash closer to 0.5.
    Returns:
        X (np.array): weakly predictive continuous features
            (num_samples, num_features)
        Y (np.array): targets (num_samples,)
    """

    cov = return_positive_semi_definite_matrix(num_features)
    X = np.random.multivariate_normal(mean=np.zeros(num_features), cov=cov, size=num_samples)

    weights = np.random.randn(num_features)
    y_probs = sigmoid(mixing_factor * np.dot(X, weights))
    y = np.random.binomial(1, p=y_probs)
    return X, y, cov, weights


def return_c_values(cardinality: int) -> Tuple[list, list]:
    """Return categorical values for C+ and C-.

    Create string values to be used for the categorical variable c.
    We build two sets of values C+ and C-.  All values from C+ end with
    "A" and all values from C- end with "B". The cardinality input
    determines len(c_pos) + len(c_neg).

    Args:
        cardinality (int): cardinality of c
    Returns:
        c_pos (list): categorical values from C+ sample
        c_neg (list): categorical values from C- sample
    """
    suffixes = [
        "{}{}".format(i, j)
        for i in string.ascii_lowercase
        for j in string.ascii_lowercase]
    c_pos = ["{}A".format(s) for s in suffixes][:int(cardinality / 2)]
    c_neg = ["{}B".format(s) for s in suffixes][:int(cardinality / 2)]
    return c_pos, c_neg


def return_strong_features(
        y_vals: np.array,
        cardinality: int,
        z_pivot: int = 10
) -> Tuple[np.array, np.array]:
    """Return strongly predictive features.

    Given a target variable values `y_vals`, create a categorical variable
    c and continuous variable z such that y is perfectly predictable from
    c and z, with y = 1 iff c takes a value from C+ OR z > z_pivot.

    Args:
        y_vals (np.array): targets
        cardinality (int): cardinality of the categorical variable, c
        z_pivot (float): mean of z
    Returns:
        c (np.array): strongly predictive categorical variable
        z (np.array): strongly predictive continuous variable
    """
    z = np.random.normal(loc=z_pivot, scale=5, size=2 * len(y_vals))
    z_pos, z_neg = z[z > z_pivot], z[z <= z_pivot]
    c_pos, c_neg = return_c_values(cardinality)
    c, z = list(), list()
    for y in y_vals:
        coin = np.random.binomial(1, 0.5)
        if y and coin:
            c.append(random.choice(c_pos + c_neg))
            z.append(random.choice(z_pos))
        elif y and not coin:
            c.append(random.choice(c_pos))
            z.append(random.choice(z_neg))
        else:
            c.append(random.choice(c_neg))
            z.append(random.choice(z_neg))
    return np.array(c), np.array(z)


def return_main_dataset(
        num_weak: int,
        num_samp: int,
        cardinality: int = 100,
        mixing_factor: float = 0.025,
) -> Tuple[pd.DataFrame, np.array, np.array]:
    """Generate training samples.

    Generate a dataset with features c and z that are perfectly predictive
    of y and additional features x_i that are weakly predictive of y and
    correlated with eachother.

    Args:
        num_weak (int): number of weakly predictive features x_i to create
        num_samp (int): number of sample to create
        cardinality (int): cardinality of the predictive categorical variable.
          half of these values will be correlated with y=1 and the other
          with y=0.
        mixing_factor (float): see `return_weak_features_and_targets`
    Returns:
        df (pd.DataFrame): dataframe with y, z, c, and x_i columns
    """
    X, y, cov, weights = return_weak_features_and_targets(num_weak, num_samp, mixing_factor)
    c, z = return_strong_features(y, cardinality)
    xcol_names = ['x{}'.format(i) for i in range(num_weak)]
    df = pd.DataFrame(X, columns=xcol_names)
    df['y'] = y
    df['z'] = z
    df['c'] = c
    df['c'] = df['c'].astype('category')
    df = df[['y', 'c', 'z'] + xcol_names]
    return df, cov, weights


def encode_as_onehot(df_main: pd.DataFrame) -> pd.DataFrame:
    """Replace string values for c with one-hot encoding."""
    df_onehot = pd.get_dummies(df_main, 'c')
    df_onehot['y'] = df_main['y'].copy()
    return df_onehot


def encode_as_int(df_main: pd.DataFrame) -> pd.DataFrame:
    """Replace string values for c with integer encoding."""
    ord_enc = OrdinalEncoder(dtype=np.int)
    c_encoded = ord_enc.fit_transform(df_main[['c']])
    df_catnum = df_main.copy()
    df_catnum['c'] = c_encoded
    df_catnum['c'] = df_catnum['c'].astype('category')
    return df_catnum, ord_enc


def encode_as_magic_int(df_main: pd.DataFrame) -> pd.DataFrame:
    """Replace string values for c with "magic" integer encoding.

    A magic encoding is one in which the sorted integer values keep all
    C+ values (values of c that end with "A") next to each other and all
    C- values (values of c that end with "B") next to eachother.
    """
    values = sorted(df_main['c'].unique(), key=lambda x: x[-1])
    ord_enc = OrdinalEncoder(categories=[values], dtype=np.int)
    c_encoded = ord_enc.fit_transform(df_main[['c']])
    df_catnum = df_main.copy()
    df_catnum['c'] = c_encoded
    df_catnum['c'] = df_catnum['c'].astype('category')
    return df_catnum, ord_enc


def get_feature_names(df, include_c=True):
    names = [f for f in df.columns if not f.startswith('y')]
    if not include_c:
        names = [f for f in names if not f.startswith('c')]
    return names


def print_auc_mean_std(results):
    print("    AUC: mean={:4.4f}, sd={:4.4f}".format(
        np.mean(results['metric']), np.std(results['metric'])))


def print_sorted_mean_importances(results, n=5):
    data = defaultdict(list)
    imps = results['importances']
    for d in imps:
        for fname, imp in d.items():
            data[fname].append(imp)
    mu = {fname: np.mean(vals) for fname, vals in data.items()}
    mu = sorted(mu.items(), key=itemgetter(1), reverse=True)[:n]
    print("    Importances:")
    for fname, val in mu:
        print("{:>20}: {:0.03f}".format(fname, val))


# @jit(nopython=True)
def powerset(iterable):
    s = list(iterable)
    return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s) + 1))


# @jit(nopython=True)
def get_partition(leaf_id, part, node_id, children_left, children_right, feature, threshold):
    left = np.where(children_left == leaf_id)[0]
    right = np.where(children_right == leaf_id)[0]

    if (len(left) == 0) * (len(right) == 0):
        return part, node_id

    else:
        if len(right) != 0:
            right = int(right[0])
            node_id.append(feature[right])

            part[feature[right]] = np.concatenate((part[feature[right]], np.array([[threshold[right], np.inf]])))
            part[feature[right]] = np.array([[np.max(part[feature[right]][:, 0]), np.min(part[feature[right]][:, 1])]])
            return get_partition(right, part, node_id, children_left, children_right, feature, threshold)
        else:
            left = int(left[0])
            node_id.append(feature[left])

            part[feature[left]] = np.concatenate((part[feature[left]], np.array([[-np.inf, threshold[left]]])))
            part[feature[left]] = np.array([[np.max(part[feature[left]][:, 0]), np.min(part[feature[left]][:, 1])]])

            return get_partition(left, part, node_id, children_left, children_right, feature, threshold)


def explore_partition(i, x, children_left, children_right, features, thresholds, values,
                      compatible_leaves, partition_leaves, partition_global, prob_global, s_global, S, S_bar, data,
                      down_tx, up_tx, intv=False):
    if children_left[i] < 0:
        #         tab[i] = 1
        compatible_leaves.append(i)
        partition_global[i] = partition_leaves
        partition_leaves = np.squeeze(np.array(partition_leaves))

        section_x = np.prod(
            [(data[:, s] <= partition_leaves[s, 1]) * (data[:, s] >= partition_leaves[s, 0]) for s in
             S], axis=0)

        section_x_bar = np.prod(
            [(data[:, s] <= partition_leaves[s, 1]) * (data[:, s] >= partition_leaves[s, 0]) for s in
             S_bar], axis=0)

        section_up = np.prod(
            [(data[:, s] <= partition_leaves[s, 1]) * (data[:, s] >= partition_leaves[s, 0]) for s in
             S], axis=0) * up_tx

        section_down = np.prod(
            [(data[:, s] <= partition_leaves[s, 1]) * (data[:, s] >= partition_leaves[s, 0]) for s in
             S], axis=0) * down_tx

        prob_all = section_x * section_x_bar
        prob_up = section_up * section_x_bar
        prob_down = section_down * section_x_bar

        s_all = section_x
        s_up = section_up
        s_down = section_down

        prob_global['all'].append(prob_all.reshape(1, -1))
        prob_global['up'].append(prob_up.reshape(1, -1))
        prob_global['down'].append(prob_down.reshape(1, -1))

        s_global['all'].append(s_all.reshape(1, -1))
        s_global['up'].append(s_up.reshape(1, -1))
        s_global['down'].append(s_down.reshape(1, -1))

    else:
        if features[i] in S:
            if x[features[i]] <= thresholds[i]:
                part_left = partition_leaves.copy()
                part_left[features[i]] = np.concatenate((part_left[features[i]], np.array([[-np.inf, thresholds[i]]])))
                part_left[features[i]] = np.array(
                    [[np.max(part_left[features[i]][:, 0]), np.min(part_left[features[i]][:, 1])]])
                explore_partition(children_left[i], x, children_left, children_right, features, thresholds, values,
                                  compatible_leaves, part_left, partition_global, prob_global, s_global, S, S_bar, data,
                                  down_tx, up_tx, intv)
            else:
                part_right = partition_leaves.copy()
                part_right[features[i]] = np.concatenate((part_right[features[i]], np.array([[thresholds[i], np.inf]])))
                part_right[features[i]] = np.array(
                    [[np.max(part_right[features[i]][:, 0]), np.min(part_right[features[i]][:, 1])]])
                explore_partition(children_right[i], x, children_left, children_right, features, thresholds, values,
                                  compatible_leaves, part_right, partition_global, prob_global, s_global, S, S_bar,
                                  data, down_tx, up_tx, intv)
        else:
            part_left = partition_leaves.copy()
            part_left[features[i]] = np.concatenate((part_left[features[i]], np.array([[-np.inf, thresholds[i]]])))
            part_left[features[i]] = np.array(
                [[np.max(part_left[features[i]][:, 0]), np.min(part_left[features[i]][:, 1])]])

            part_right = partition_leaves.copy()
            part_right[features[i]] = np.concatenate((part_right[features[i]], np.array([[thresholds[i], np.inf]])))
            part_right[features[i]] = np.array(
                [[np.max(part_right[features[i]][:, 0]), np.min(part_right[features[i]][:, 1])]])

            explore_partition(children_left[i], x, children_left, children_right, features, thresholds, values,
                              compatible_leaves, part_left, partition_global, prob_global, s_global, S, S_bar, data
                              , down_tx, up_tx, intv)
            explore_partition(children_right[i], x, children_left, children_right, features, thresholds, values,
                              compatible_leaves, part_right, partition_global, prob_global, s_global, S, S_bar, data
                              , down_tx, up_tx, intv)


def get_tree_partition(x, fx, tx, tree, S, data=None, is_reg=True):
    """
    Compute the partition (L_m) of each compatible leaf of the condition X_s = x_S, then check for each
    observations in data in which leaves it falls.

    Args:
        x (array): observation
        fx (float): tree(x)
        tx (float): threshold of the classifier
        tree (DecisionTreeClassifier.tree_): model
        S (list): index of variables on which we want to compute the SDP
        algo (string): name of the estimators, recommended 'pluging'
        data (array): data used to compute the partion

    Returns:
        (array, array): binary array of shape (data_size, compatible_leaves), if [i, j] = 1 then observation i fall in
        leaf j.
        (array): return number of observations that fall in each leaf
    """

    children_left = tree.children_left
    children_right = tree.children_right
    features = tree.features
    thresholds = tree.thresholds
    # r_w = tree.node_samples_weight
    index = range(x.shape[0])
    if is_reg:
        values = tree.values.reshape(-1) / tree.scaling
        y_pred = tree.predict(data)
        dist = (y_pred - fx) ** 2

        up_tx = np.array(dist > tx).reshape(-1)
        down_tx = np.array(dist <= tx).reshape(-1)
    else:
        values = tree.values / tree.scaling
        y_pred = tree.predict(data)

        if len(y_pred.shape) == 1:
            y_pred = np.array([1 - y_pred, y_pred]).T

        argmax_y_pred = np.argmax(y_pred, axis=1)

        up_tx = np.array(argmax_y_pred == int(fx)).reshape(-1)
        down_tx = np.array(argmax_y_pred != int(fx)).reshape(-1)

    S_bar = [i for i in index if i not in S]
    partition_leaves = [np.array([[-np.inf, np.inf]]) for i in range(data.shape[1])]
    partition_global = {i: [np.array([[-np.inf, np.inf]]) for i in range(data.shape[1])]
                        for i in range(len(tree.features))}

    prob_global = {'all': [], 'up': [], 'down': []}
    s_global = {'all': [], 'up': [], 'down': []}

    part_final = {}
    compatible_leaves = []
    explore_partition(0, x, children_left, children_right, features, thresholds, values,
                      compatible_leaves, partition_leaves, partition_global, prob_global, s_global, S, S_bar, data,
                      down_tx, up_tx, intv=False)

    part_final['all'] = np.concatenate(prob_global['all'], axis=0)
    part_final['up'] = np.concatenate(prob_global['up'], axis=0)
    part_final['down'] = np.concatenate(prob_global['down'], axis=0)

    part_final['s_all'] = np.concatenate(s_global['all'], axis=0)
    part_final['s_up'] = np.concatenate(s_global['up'], axis=0)
    part_final['s_down'] = np.concatenate(s_global['down'], axis=0)

    return part_final, values[compatible_leaves]


import_errors = {}


def assert_import(package_name):
    global import_errors
    if package_name in import_errors:
        msg, e = import_errors[package_name]
        print(msg)
        raise e


def record_import_error(package_name, msg, e):
    global import_errors
    import_errors[package_name] = (msg, e)


def safe_isinstance(obj, class_path_str):
    """
    Acts as a safe version of isinstance without having to explicitly
    import packages which may not exist in the users environment.

    Checks if obj is an instance of type specified by class_path_str.

    Parameters
    ----------
    obj: Any
        Some object you want to test against
    class_path_str: str or list
        A string or list of strings specifying full class paths
        Example: `sklearn.ensemble.RandomForestRegressor`

    Returns
    --------
    bool: True if isinstance is true and the package exists, False otherwise
    """
    if isinstance(class_path_str, str):
        class_path_strs = [class_path_str]
    elif isinstance(class_path_str, list) or isinstance(class_path_str, tuple):
        class_path_strs = class_path_str
    else:
        class_path_strs = ['']

    # try each module path in order
    for class_path_str in class_path_strs:
        if "." not in class_path_str:
            raise ValueError("class_path_str must be a string or list of strings specifying a full \
                module path to a class. Eg, 'sklearn.ensemble.RandomForestRegressor'")

        # Splits on last occurence of "."
        module_name, class_name = class_path_str.rsplit(".", 1)

        # here we don't check further if the model is not imported, since we shouldn't have
        # an object of that types passed to us if the model the type is from has never been
        # imported. (and we don't want to import lots of new modules for no reason)
        if module_name not in sys.modules:
            continue

        module = sys.modules[module_name]

        # Get class
        _class = getattr(module, class_name, None)

        if _class is None:
            continue

        if isinstance(obj, _class):
            return True

    return False


def get_null_coalition(s_star, len_s_star):
    n_star = -np.ones(s_star.shape, dtype=np.long)
    index = list(range(s_star.shape[1]))

    for i in range(s_star.shape[0]):
        s_star_index = [s_star[i, j] for j in range(s_star.shape[1])]
        null_coalition = list(set(index) - set(s_star_index))
        n_star[i, len_s_star[i]:] = np.array(null_coalition)
    return s_star, n_star


def get_active_null_coalition_list(s_star, len_s_star):
    index = list(range(s_star.shape[1]))
    s_star_all = []
    n_star_all = []
    for i in range(s_star.shape[0]):
        s_star_all.append([s_star[i, j] for j in range(len_s_star[i])])
        n_star_all.append(list(set(index) - set(s_star_all[-1])))
    return s_star_all, n_star_all


class ModelW:
    def __init__(self, model, prediction='predict_proba'):
        self.model = model
        self.prediction = prediction

    def __call__(self, x):
        if self.prediction == 'predict_proba':
            if len(x.shape) == 1:
                return self.model.predict_proba(x.reshape(-1, 1))
            return self.model.predict_proba(x)
        elif self.prediction == 'predict_proba_one':
            if len(x.shape) == 1:
                return self.model.predict_proba(x.reshape(-1, 1))[:, 1]
            return self.model.predict_proba(x)[:, 1]
        else:
            if len(x.shape) == 1:
                return self.model.predict(x.reshape(-1, 1))
            return self.model.predict(x)

    def predict(self, x):
        return self.__call__(x)


def weighted_percentile(a, q, weights=None, sorter=None):
    """
    Returns the weighted percentile of a at q given weights.
    Parameters
    ----------
    a: array-like, shape=(n_samples,)
        samples at which the quantile.
    q: int
        quantile.
    weights: array-like, shape=(n_samples,)
        weights[i] is the weight given to point a[i] while computing the
        quantile. If weights[i] is zero, a[i] is simply ignored during the
        percentile computation.
    sorter: array-like, shape=(n_samples,)
        If provided, assume that a[sorter] is sorted.
    Returns
    -------
    percentile: float
        Weighted percentile of a at q.
    References
    ----------
    1. https://en.wikipedia.org/wiki/Percentile#The_Weighted_Percentile_method
    Notes
    -----
    Note that weighted_percentile(a, q) is not equivalent to
    np.percentile(a, q). This is because in np.percentile
    sorted(a)[i] is assumed to be at quantile 0.0, while here we assume
    sorted(a)[i] is given a weight of 1.0 / len(a), hence it is at the
    1.0 / len(a)th quantile.
    """
    if weights is None:
        weights = np.ones_like(a)
    if q > 100 or q < 0:
        raise ValueError("q should be in-between 0 and 100, "
                         "got %d" % q)

    a = np.asarray(a, dtype=np.float32)
    weights = np.asarray(weights, dtype=np.float32)
    if len(a) != len(weights):
        raise ValueError("a and weights should have the same length.")

    if sorter is not None:
        a = a[sorter]
        weights = weights[sorter]

    nz = weights != 0
    a = a[nz]
    weights = weights[nz]

    if sorter is None:
        sorted_indices = np.argsort(a)
        sorted_a = a[sorted_indices]
        sorted_weights = weights[sorted_indices]
    else:
        sorted_a = a
        sorted_weights = weights

    # Step 1
    sorted_cum_weights = np.cumsum(sorted_weights)
    total = sorted_cum_weights[-1]

    # Step 2
    partial_sum = 100.0 / total * (sorted_cum_weights - sorted_weights / 2.0)
    start = np.searchsorted(partial_sum, q) - 1
    if start == len(sorted_cum_weights) - 1:
        return sorted_a[-1]
    if start == -1:
        return sorted_a[0]

    # Step 3.
    fraction = (q - partial_sum[start]) / (partial_sum[start + 1] - partial_sum[start])
    return sorted_a[start] + fraction * (sorted_a[start + 1] - sorted_a[start])


def find_nbor(rec_a, rec_b, S):
    axs = []
    dim = []
    for k in S:
        if rec_a[k, 0] == rec_b[k, 1]:
            axs.append(k)
            dim.append(0)
        elif rec_a[k, 1] == rec_b[k, 0]:
            axs.append(k)
            dim.append(1)
    return axs, dim


def extend_rec(rec_a, rec_b, S, axs, dim):
    a = 0
    for k in S:
        if k not in axs:
            if not rec_b[k, 0] <= rec_a[k, 0] and rec_b[k, 1] >= rec_a[k, 1]:
                a += 1
    if a == 0:
        for i, k in enumerate(axs):
            rec_a[k, dim[i]] = rec_b[k, dim[i]]
    return rec_a


def find_union(rec_a, list_ric, S):
    axs, dim = [], []
    for i, rec_b in enumerate(list_ric):
        axs, dim = find_nbor(rec_a, rec_b, S)
        if len(axs) != 0:
            break
    if len(axs) == 0 or len(list_ric) == 0:
        return rec_a
    else:
        del list_ric[i]
        rec_a = extend_rec(rec_a, rec_b, S, axs, dim)
        return find_union(rec_a, list_ric, S)


def extend_partition(rules, rules_data, sdp_all, pi, S):
    for i in range(rules.shape[0]):
        list_ric = [rules_data[i, j] for j in range(rules_data.shape[1]) if sdp_all[i, j] >= pi]
        find_union(rules[i], list_ric, S=S[i])


def global_rules_model(x_exp, rules, rules_output, rules_coverage, rules_acc, min_cov, S, min_acc=0.8):
    y_exp, rule_exp, rule_S = [], [], []
    for j in range(x_exp.shape[0]):
        x = x_exp[j]

        y_out, y_coverage, y_weights, y_rule, y_S = [], [], [], [], []
        for i in range(rules.shape[0]):
            rule = rules[i]

            x_in = np.prod([(x[s] <= rule[s, 1]) * (x[s] >= rule[s, 0]) for s in S[i]], axis=0).astype(bool)
            if x_in:
                y_out.append(rules_output[i])
                y_coverage.append(rules_coverage[i])
                y_weights.append(rules_acc[i])
                y_rule.append(rule)
                y_S.append(S[i])

        y_weights = np.array(y_weights)
        if len(y_out) != 0 and np.max(y_weights[np.array(y_coverage) >= min_cov]) >= min_acc:
            max_weights = np.max(y_weights[np.array(y_coverage) >= min_cov])
            best_acc = np.argmax(y_weights == max_weights)

            y_exp.append(y_out[best_acc])
            rule_exp.append(y_rule[best_acc])
            rule_S.append(y_S[best_acc])

        else:
            y_exp.append(None)
            rule_exp.append(None)
            rule_S.append(None)

    return y_exp, rule_exp, rule_S


def global_rules_model_reg(x_exp, rules, rules_output, rules_coverage, rules_acc, min_cov, S, min_mse=500):
    y_exp, rule_exp, rule_S = [], [], []
    for j in range(x_exp.shape[0]):
        x = x_exp[j]

        y_out, y_coverage, y_weights, y_rule, y_S = [], [], [], [], []
        for i in range(rules.shape[0]):
            rule = rules[i]

            x_in = np.prod([(x[s] <= rule[s, 1]) * (x[s] >= rule[s, 0]) for s in S[i]], axis=0).astype(bool)
            if x_in:
                y_out.append(rules_output[i])
                y_coverage.append(rules_coverage[i])
                y_weights.append(rules_acc[i])
                y_rule.append(rule)
                y_S.append(S[i])

        y_weights = np.array(y_weights)
        if len(y_out) != 0 and np.min(y_weights[np.array(y_coverage) >= min_cov]) <= min_mse:
            min_weights = np.min(y_weights[np.array(y_coverage) >= min_cov])
            best_acc = np.argmax(y_weights == min_weights)

            y_exp.append(y_out[best_acc])
            rule_exp.append(y_rule[best_acc])
            rule_S.append(y_S[best_acc])

        else:
            y_exp.append(None)
            rule_exp.append(None)
            rule_S.append(None)

    return y_exp, rule_exp, rule_S


def compute_rules_metrics(rules, data, y_data, S_star, classifier=True):
    d = data.shape[1]
    rules_coverage = []
    rules_var = []
    rules_acc = []
    rules_output = []
    rules_output_proba = []
    for idx in range(rules.shape[0]):
        S = S_star[idx]
        rule = rules[idx]
        where = np.prod([(data[:, s] <= rule[s, 1]) * (data[:, s] >= rule[s, 0])
                         for s in S], axis=0).astype(bool)
        if classifier:
            rules_output.append(1 * (np.mean(y_data[where]) > 0.5))
            rules_output_proba.append((np.mean(y_data[where])))
            rules_acc.append(np.mean(y_data[where] == rules_output[-1]))
        else:
            rules_output.append(np.mean(y_data[where]))
            rules_acc.append(np.mean((y_data[where] - rules_output[-1]) ** 2))

        rules_coverage.append(np.sum(where) / data.shape[0])
        rules_var.append(np.var(y_data[where]))
    if classifier:
        return rules_coverage, rules_acc, rules_var, rules_output, rules_output_proba
    return rules_coverage, rules_acc, rules_var, rules_output


def rules_frequency(rules):
    freq = np.zeros(rules.shape[0])
    for i in range(rules.shape[0]):
        rule = rules[i]
        for j in range(rules.shape[0]):
            if np.allclose(rules[j], rule):
                freq[i] += 1
    freq = freq / rules.shape[0]
    return freq


def unique_rules(rules, rules_output):
    rules_unique = np.unique(rules, axis=0)
    rules_unique_output = []
    for i in range(rules_unique.shape[0]):
        rule = rules_unique[i]
        for j in range(rules.shape[0]):
            if np.allclose(rule, rules[j]):
                rules_unique_output.append(rules_output[j])
                break
    return rules_unique, np.array(rules_unique_output)
